"""
mdb.writeable: subclass of `class:bento_meta.MDB` to support writing to an MDB
"""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Any, Concatenate, ParamSpec, cast

from typing_extensions import LiteralString

from bento_meta.mdb.mdb import MDB, make_nanoid

if TYPE_CHECKING:
    from collections.abc import Callable

    from neo4j import ManagedTransaction, Record

    from bento_meta.objects import Term


P = ParamSpec("P")


def write_txn(
    func: Callable[Concatenate[WriteableMDB, P], tuple[str, dict[str, Any] | None]],
) -> Callable[Concatenate[WriteableMDB, P], list[Record]]:
    """
    Decorate a query function to run a write transaction based on its query.

    Query function should return a tuple (qry_string, param_dict).

    Args:
        func: The query function to decorate.

    Returns:
        Decorated function that executes a write transaction.
    """

    @wraps(func)
    def wr(self: WriteableMDB, *args: P.args, **kwargs: P.kwargs) -> list[Record]:
        def txn_q(tx: ManagedTransaction) -> list[Record]:
            (qry, parms) = func(self, *args, **kwargs)
            result = tx.run(cast("LiteralString", qry), parameters=parms)
            return list(result)

        with self.driver.session() as session:
            return session.execute_write(txn_q)

    return wr


class WriteableMDB(MDB):
    """:class:`bento_meta.mdb.MDB` subclass for writing to an MDB."""

    def __init__(
        self,
        uri: str | None = None,
        user: str | None = None,
        password: str | None = None,
    ) -> None:
        """Initialize a :class:`WriteableMDB` object."""
        super().__init__(uri=uri, user=user, password=password)

    @write_txn  # type: ignore[reportArgumentType]
    def put_with_statement(
        self,
        qry: str,
        parms: dict[str, Any] | None = None,
    ) -> list[Record]:
        """Run an arbitrary write statement."""
        if parms is None:
            parms = {}
        if not isinstance(qry, str):
            msg = "qry= must be a string"
            raise TypeError(msg)
        if not isinstance(parms, dict):
            msg = "parms= must be a dict"
            raise TypeError(msg)
        return (qry, parms)  # type: ignore[reportReturnType]

    @write_txn  # type: ignore[reportArgumentType]
    def put_term_with_origin(
        self,
        term: Term,
        commit: str = "",
        _from: int = 1,
    ) -> list[Record]:
        """
        Merge a bento-meta Term object, that has an Origin object set into an MDB.

        If a new term is created, assign a random 6-char nanoid to it.
        The Origin must already be represented in the database.

        Args:
            term: Term object to merge.
            commit: GitHub commit SHA1 associated with the term (if any).
            _from: Source identifier.

        Returns:
            List of Records from the transaction.
        """
        qry = (
            "match (o:origin {name:$o_name, nanoid:$o_id}) "
            "where not exists(o._to) "
            "merge (t:term {"
            "  value:$t_value,"
            "  origin_id:$t_oid,"
            "  origin_version:$t_oversion,"
            "  origin_definition:$t_odefn,"
            "  _from:$from, _commit:$commit}) "
            "merge (o)<-[:has_origin]-(t) "
            "on create set t.nanoid = $t_id "
            "return t.nanoid"
        )
        parms = {
            "o_name": term.origin.name,
            "o_id": term.origin.nanoid,
            "t_value": term.value,
            "t_oid": term.origin_id,
            "t_oversion": term.origin_version,
            "t_odefn": term.origin_definition,
            "t_id": make_nanoid(),
            "commit": commit,
            "from": _from,
        }
        return (qry, parms)  # type: ignore[reportReturnType]



"""
bento_meta.objects
==================

This module contains the subclasses of :class:`Entity` which are used
in representing the models contained in the `MDB <https://github.com/CBIIT/bento-mdf>`_.

"""

from __future__ import annotations

import sys

sys.path.append("..")
from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar
from warnings import warn

from bento_meta.entity import Entity

if TYPE_CHECKING:
    import neo4j


def mergespec(
    clsname: str,
    attspec: dict[str, str],
    mapspec: dict[
        str,
        str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None,
    ],
) -> tuple[
    dict[str, str],
    dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None],
]:
    """
    Merge subclass attribute and mapping specification dicts with the base class's.

    Not for human consumption.
    """
    spec = deepcopy(attspec)
    spec.update(Entity.attspec_)
    mo = deepcopy(Entity.mapspec_)
    if "label" in mapspec:
        mo["label"] = mapspec["label"]
    if "key" in mapspec:
        mo["key"] = mapspec["key"]
    if "property" in mapspec:
        mo["property"].update(mapspec["property"])
    if "relationship" in mapspec:
        mo["relationship"].update(mapspec["relationship"])
    mo["relationship"]["_next"]["end_cls"] = {clsname}
    mo["relationship"]["_prev"]["end_cls"] = {clsname}
    return (spec, mo)


class Node(Entity):
    """Subclass that models a data node."""

    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "model": "simple",
        "nanoid": "simple",
        "version": "simple",
        "concept": "object",
        "props": "collection",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "node",
        "key": "handle",
        "property": {
            "handle": "handle",
            "model": "model",
            "nanoid": "nanoid",
            "version": "version",
        },
        "relationship": {
            "concept": {"rel": ":has_concept>", "end_cls": "Concept"},
            "props": {"rel": ":has_property>", "end_cls": "Property"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("Node", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Node | None = None) -> None:
        """Initialize a `Node` instance."""
        super().__init__(init=init)

    @property
    def annotations(self) -> dict | None:
        """Return `Term`s if the `Node` is annotated by `Term`s via a `Concept`."""
        if self.concept:
            return self.concept.terms
        return None

    def get_key_prop(self) -> Property | list[Property] | None:
        """
        Return the Property entity with is_key=True for this Node if it exists.

        If multiple key props exist, return a list of them; if none exist, return None.

        Returns:
            Single Property if one key property exists, list of Properties if multiple
            exist, None if no key properties found.
        """
        if not self.props.values():
            warn("No properties found for Node - returning None", stacklevel=2)
            return None
        keys = [p for p in self.props.values() if p.is_key]
        if not keys:
            warn("No key properties found for Node - returning None", stacklevel=2)
            return None
        if len(keys) == 1:
            return keys[0]
        warn("Multiple key properties found for Node - returning all", stacklevel=2)
        return keys


class Property(Entity):
    """Subclass that models a property of a node or relationship (edge)."""

    pvt_attr: ClassVar[list[str]] = [*Entity.pvt_attr, "value_types"]
    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "model": "simple",
        "nanoid": "simple",
        "version": "simple",
        "value_domain": "simple",
        "units": "simple",
        "pattern": "simple",
        "item_domain": "simple",
        "is_required": "simple",
        "is_key": "simple",
        "is_nullable": "simple",
        "is_deprecated": "simple",
        "is_strict": "simple",
        "concept": "object",
        "value_set": "object",
        "_parent_handle": "simple",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "property",
        "key": "handle",
        "property": {
            "handle": "handle",
            "model": "model",
            "nanoid": "nanoid",
            "value_domain": "value_domain",
            "pattern": "pattern",
            "units": "units",
            "item_domain": "item_domain",
            "is_required": "is_required",
            "is_key": "is_key",
            "is_nullable": "is_nullable",
            "is_deprecated": "is_deprecated",
            "is_strict": "is_strict",
            "_parent_handle": "_parent_handle",
            "version": "version",
        },
        "relationship": {
            "concept": {"rel": ":has_concept>", "end_cls": "Concept"},
            "value_set": {"rel": ":has_value_set>", "end_cls": "ValueSet"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("Property", attspec_, mapspec_)
    defaults: ClassVar[dict[str, str]] = {"value_domain": "TBD"}

    def __init__(self, init: dict | neo4j.graph.Node | Property | None = None) -> None:
        """Initialize a `Property` instance."""
        super().__init__(init=init)
        self.value_types = []

    @property
    def annotations(self) -> dict | None:
        """If the `Property` is annotated by `Term`s via a `Concept`, return the `Term`s."""
        if self.concept:
            return self.concept.terms
        return None

    @property
    def terms(self) -> list[Term] | None:
        """Return the `Term` objects of `Property` with a ``value_set`` domain."""
        if self.value_set:
            return self.value_set.terms
        return None

    @property
    def values(self) -> list[str] | None:
        """
        Return Property's term values as a list of str if it has a value_set domain.

        Returns:
            List of term values, or None if no value_set.
        """
        if self.value_set:
            return [self.terms[x].value for x in self.terms]
        return None


class Edge(Entity):
    """Subclass that models a relationship between model nodes."""

    defaults: ClassVar[dict[str, str]] = {
        "multiplicity": "many_to_many",
    }
    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "model": "simple",
        "nanoid": "simple",
        "version": "simple",
        "multiplicity": "simple",
        "is_required": "simple",
        "src": "object",
        "dst": "object",
        "concept": "object",
        "props": "collection",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "relationship",
        "key": "handle",
        "property": {
            "handle": "handle",
            "model": "model",
            "nanoid": "nanoid",
            "multiplicity": "multiplicity",
            "is_required": "is_required",
            "version": "version",
        },
        "relationship": {
            "src": {"rel": ":has_src>", "end_cls": "Node"},
            "dst": {"rel": ":has_dst>", "end_cls": "Node"},
            "concept": {"rel": ":has_concept>", "end_cls": "Concept"},
            "props": {"rel": ":has_property>", "end_cls": "Property"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("Edge", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Edge | None = None) -> None:
        """Initialize an `Edge` instance."""
        super().__init__(init=init)

    @property
    def annotations(self) -> dict | None:
        """Return `Term`s if the `Edge` is annotated by `Term`s via a `Concept`."""
        if self.concept:
            return self.concept.terms
        return None

    @property
    def triplet(self) -> tuple[str, str, str] | None:
        """
        Return a 3-tuple that fully qualifies the edge.

        Returns (edge.handle, src.handle, dst.handle).

        src and dst attributes must be set. If not, return None.

        Returns:
            Tuple of (edge.handle, src.handle, dst.handle) or None.
        """
        if self.handle and self.src and self.dst:
            return (self.handle, self.src.handle, self.dst.handle)
        return None


class Term(Entity):
    """Subclass that models a term from a terminology."""

    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "value": "simple",
        "nanoid": "simple",
        "origin_id": "simple",
        "origin_version": "simple",
        "origin_definition": "simple",
        "origin_name": "simple",
        "concept": "object",
        "origin": "object",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "term",
        # note using 'value' as the key in terms collections may break silently
        # if the terms are coming from different origins, but use the same
        # value (string representation in data) - i.e., value is not necessarily
        # unique - should be the nanoid of the term.
        "key": "value",
        "property": {
            "handle": "handle",
            "value": "value",
            "nanoid": "nanoid",
            "origin_id": "origin_id",
            "origin_version": "origin_version",
            "origin_definition": "origin_definition",
            "origin_name": "origin_name",
        },
        "relationship": {
            "concept": {"rel": ":represents>", "end_cls": "Concept"},
            "origin": {"rel": ":has_origin>", "end_cls": "Origin"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("Term", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Term | None = None) -> None:
        """Initialize a `Term` instance."""
        super().__init__(init=init)


# for ValueSet - updating terms prop should dirty the connected Property
# (from Bento::Meta), signal need to refresh. Engineer so this happens
# here (__setattr__ override), not in Entity
class ValueSet(Entity):
    """
    Subclass that models an enumerated set of Property values.

    Essentially a container for Term instances.
    """

    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "nanoid": "simple",
        "url": "simple",
        "path": "simple",
        "prop": "object",
        "origin": "object",
        "terms": "collection",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "value_set",
        "property": {
            "handle": "handle",
            "url": "url",
            "path": "path",
            "nanoid": "nanoid",
        },
        "relationship": {
            "prop": {"rel": "<:has_value_set", "end_cls": "Property"},
            "terms": {"rel": ":has_term>", "end_cls": "Term"},
            "origin": {"rel": ":has_origin>", "end_cls": "Origin"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("ValueSet", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | ValueSet | None = None) -> None:
        """Initialize a `ValueSet` instance."""
        super().__init__(init=init)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if name == "dirty" and self.prop and value != 0:
            self.prop.dirty = 1


class Concept(Entity):
    """Subclass that models a semantic concept."""

    attspec_: ClassVar[dict[str, str]] = {"terms": "collection"}
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "concept",
        "relationship": {
            "terms": {"rel": "<:represents", "end_cls": "Term"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("Concept", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Concept | None = None) -> None:
        """Initialize a `Concept` instance."""
        super().__init__(init=init)


class Predicate(Entity):
    """Subclass that models a semantic link between concepts."""

    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "subject": "object",
        "object": "object",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "predicate",
        "key": "handle",
        "property": {
            "handle": "handle",
        },
        "relationship": {
            "subject": {"rel": ":has_subject>", "end_cls": "Concept"},
            "object": {"rel": ":has_object>", "end_cls": "Concept"},
            "tags": {"rel": ":has_tag>", "end_cls": "Tag"},
        },
    }
    (attspec, _mapspec) = mergespec("Predicate", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Predicate | None = None) -> None:
        """Initialize a `Predicate` instance."""
        super().__init__(init=init)


class Origin(Entity):
    """Subclass that models a :class:`Term` 's authoritative source."""

    attspec_: ClassVar[dict[str, str]] = {
        "url": "simple",
        "is_external": "simple",
        "name": "simple",
        "nanoid": "simple",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "origin",
        "key": "name",
        "property": {
            "name": "name",
            "url": "url",
            "is_external": "is_external",
            "nanoid": "nanoid",
        },
        "relationship": {"tags": {"rel": ":has_tag>", "end_cls": "Tag"}},
    }
    (attspec, _mapspec) = mergespec("Origin", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Origin | None = None) -> None:
        """Initialize an `Origin` instance."""
        super().__init__(init=init)


class Tag(Entity):
    """Subclass that allows simple key-value tagging of a model at arbitrary points."""

    attspec_: ClassVar[dict[str, str]] = {
        "key": "simple",
        "value": "simple",
        "_parent": "object",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "tag",
        "key": "key",
        "property": {"key": "key", "value": "value"},
        "relationship": {"_parent": {"rel": "<:has_tag",
                                     "end_cls": ["Node", "Edge", "Property", "ValueSet","Concept","Predicate","Transform","TfStep"]}},
    }
    (attspec, _mapspec) = mergespec("Tag", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Tag | None = None) -> None:
        """Initialize a `Tag` instance."""
        super().__init__(init=init)


class Model(Entity):
    """Subclass with information regarding data model."""

    attspec_: ClassVar[dict[str, str]] = {
        "handle": "simple",
        "name": "simple",
        "repository": "simple",
        "nanoid": "simple",
        "version": "simple",
        "is_latest_version": "simple",
    }
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": "model",
        "key": "handle",
        "property": {
            "handle": "handle",
            "name": "name",
            "repository": "repository",
            "version": "version",
            "is_latest_version": "is_latest_version",
        },
    }
    (attspec, _mapspec) = mergespec("Model", attspec_, mapspec_)
    defaults: ClassVar[dict[str, bool]] = {"is_latest_version": False}

    def __init__(self, init: dict | neo4j.graph.Node | Model | None = None) -> None:
        """Initialize a `Model` instance."""
        super().__init__(init=init)



"""
bento_meta.model
================

This module contains :class:`Model`, a class for managing data models housed
in the Bento Metamodel Database. Models are built from `bento_meta.Entity`
subclasses (see :mod:`bento_meta.objects`). A Model can be used with or
without a Neo4j database connection.

"""

from __future__ import annotations

import re
import sys

sys.path.append("..")
import builtins
import contextlib
from uuid import uuid4
from warnings import warn

import neo4j.graph

from bento_meta.entity import ArgError, Entity
from bento_meta.mdb import MDB, make_nanoid
from bento_meta.object_map import ObjectMap
from bento_meta.objects import (
    Concept,
    Edge,
    Node,
    Origin,
    Predicate,
    Property,
    Tag,
    Term,
    ValueSet,
)


class Model:
    """Model class for managing data models housed in the Bento Metamodel Database."""

    def __init__(
        self,
        handle: str | None = None,
        version: str | None = None,
        uri: str | None = None,
        mdb: MDB | None = None,
    ) -> None:
        """
        Model constructor.

        Args:
            handle: A string name for the model. Corresponds to the model property
                in MDB database nodes.
            version: Version string for the model.
            uri: URI for the model.
            mdb: An MDB object containing the db connection.
        """
        if not handle:
            msg = "model requires arg 'handle' set"
            raise ArgError(msg)
        self.handle = handle
        self.version = version
        self.uri = uri
        self.repository = None
        self._mdb: MDB | None = None
        self.nodes: dict[str, Node] = {}
        self.edges: dict[
            tuple[str, str, str],
            Edge,
        ] = {}  # keys are (edge.handle, src.handle, dst.handle) tuples
        self.props: dict[
            tuple[str, str],
            Property,
        ] = {}  # keys are ({edge|node}.handle, prop.handle) tuples
        self.terms: dict[
            tuple[str, str, str | None, str | None],
            Term,
        ] = {}  # keys are (term.handle, term.origin_name, term.origin_id, term.origin_version) tuples
        self.removed_entities: list[Entity] = []

        if mdb:
            self.mdb = mdb

    @property
    def drv(self) -> neo4j.Driver | None:
        """Neo4j database driver from MDB object."""
        return self._mdb.driver if self._mdb else None

    @property
    def mdb(self) -> MDB | None:
        """MDB object containing the db connection."""
        return self._mdb

    @mdb.setter
    def mdb(self, value: MDB | None) -> None:
        """Set the MDB object containing the db connection."""
        if isinstance(value, MDB):
            self._mdb = value
            for cls in (
                Node,
                Property,
                Edge,
                Term,
                ValueSet,
                Concept,
                Predicate,
                Origin,
                Tag,
            ):
                cls.object_map = ObjectMap(cls=cls, drv=value.driver)
        elif not value:
            self._mdb = None
            for cls in (Node, Property, Edge, Term, ValueSet, Concept, Origin, Tag):
                cls.object_map = None
        else:
            msg = "mdb= arg must be a bento_meta.mdb.MDB object"
            raise ArgError(msg)

    def add_node(self, node: Node | dict | neo4j.graph.Node | None = None) -> Node:
        """
        Add a Node to the model.

        The model attribute of node is set to Model.handle.

        Args:
            node: A Node instance, a neo4j.graph.Node, or a dict.

        Returns:
            The added Node instance.
        """
        if not node:
            msg = "arg must be Node, dict, or graph.Node"
            raise ArgError(msg)
        if isinstance(node, (dict, neo4j.graph.Node)):
            node = Node(node)
        if not node.model:
            node.model = self.handle
        for p in node.props.values():
            self.add_prop(node, p)
        self.nodes[node.handle] = node
        return node

    def add_edge(self, edge: Edge | dict | neo4j.graph.Node | None = None) -> Edge:
        """
        Add an Edge to the model.

        The model attribute of edge is set to Model.handle.

        Args:
            edge: An Edge instance, a neo4j.graph.Node, or a dict.

        Returns:
            The added Edge instance.
        """
        if not edge:
            msg = "arg must be Edge, dict, or graph.Node"
            raise ArgError(msg)
        if isinstance(edge, (dict, neo4j.graph.Node)):
            edge = Edge(edge)
        if not edge.src or not edge.dst:
            msg = "edge must have both src and dst set"
            raise ArgError(msg)
        if not edge.model:
            edge.model = self.handle
        if not self.contains(edge.src):
            warn("Edge source node not yet in model; adding it", stacklevel=2)
            self.add_node(edge.src)
        if not self.contains(edge.dst):
            warn("Edge destination node not yet in model; adding it", stacklevel=2)
            self.add_node(edge.dst)
        for p in edge.props.values():
            self.add_prop(edge, p)
        self.edges[edge.triplet] = edge
        return edge

    def add_prop(
        self,
        ent: Node | Edge,
        prop: Property | dict | neo4j.graph.Node | None = None,
        *,
        reuse: bool = False,
    ) -> Property:
        """
        Add a Property to the model.

        The model attribute of prop is set to Model.handle. Within a model,
        Property entities are unique with respect to their handle (but can be reused).
        This method will look for an existing property within the model with the given
        handle, and add an item to Model.props pointing to it if found.

        Args:
            ent: Attach prop to this entity (Node or Edge).
            prop: A Property instance, a neo4j.graph.Node, or a dict.
            reuse: If True, reuse existing property with same handle.

        Returns:
            The added Property instance.
        """
        if not isinstance(ent, (Node, Edge)):
            msg = "arg 1 must be Node or Edge"
            raise ArgError(msg)
        if not prop:
            msg = "arg 2 must be Property, dict, or graph.Node"
            raise ArgError(msg)
        if isinstance(prop, (dict, neo4j.graph.Node)):
            prop = Property(prop)
        if not prop.model:
            prop.model = self.handle
        if prop.value_domain == "value_set" and not prop.value_set:
            warn(
                "(add_prop) Creating ValueSet object for Property " + prop.handle,
                stacklevel=2,
            )
            prop.value_set = ValueSet({"prop": prop, "_id": str(uuid4())})
            prop.value_set.handle = self.handle + prop.value_set._id[0:8]  # noqa: SLF001
        key = [ent.handle] if isinstance(ent, Node) else list(ent.triplet)
        key.append(prop.handle)
        ent.props[getattr(prop, type(prop).mapspec()["key"])] = prop
        if tuple(key) not in self.props:
            self.props[tuple(key)] = prop
        return prop

    def annotate(self, ent: Entity, term: Term) -> None:
        """
        Associate a single Term with an Entity.

        This creates a Concept entity if needed and links both the Entity and the Term
        to the concept, in keeping with the MDB spec. It supports the Term key in MDF.

        Args:
            ent: Entity object to annotate.
            term: Term object to describe the Entity.
        """
        if not isinstance(ent, Entity):
            msg = "arg1 must be Entity"
            raise ArgError(msg)
        if not isinstance(term, Term):
            msg = "arg2 must be Term"
            raise ArgError(msg)
        if not ent.concept:
            ent.concept = Concept({"nanoid": make_nanoid()})
        term_key = term.handle if term.handle else term.value
        full_term_key = (
            term_key,
            term.origin_name,
            term.origin_id,
            term.origin_version,
        )
        if full_term_key in ent.concept.terms:
            msg = (
                "Concept already represented by a Term with handle or value "
                f"'{term_key}', origin_name '{term.origin_name}', origin_id '{term.origin_id}', "
                f"and origin_version '{term.origin_version}'"
            )
            raise ValueError(msg)
        ent.concept.terms[full_term_key] = term
        self.terms[full_term_key] = term

    def add_terms(self, prop: Property, *terms: list[Term | str]) -> None:
        """
        Add a list of Term and/or strings to a Property.

        Property must have a value domain of value_set or enum.
        Term instances are created for strings; Term.value and Term.handle
        is set to the string.

        Args:
            prop: Property to modify.
            terms: A list of Term instances and/or str.
        """
        if not isinstance(prop, Property):
            msg = "arg1 must be Property"
            raise ArgError(msg)
        if not re.match("value_set|enum", prop.value_domain):
            msg = "Property value domain is not value_set or enum, can't add terms"
            raise AttributeError(msg)
        if not prop.value_set:
            warn(
                "(add_terms) Creating ValueSet object for Property " + prop.handle,
                stacklevel=2,
            )
            prop.value_set = ValueSet({"prop": prop, "_id": str(uuid4())})
            prop.value_set.handle = self.handle + prop.value_set._id[0:8]  # noqa: SLF001

        for item in terms:
            if isinstance(item, str):
                warn(f"Creating Term object for string '{item}'", stacklevel=2)
                term = Term({"handle": item, "value": item})
            elif isinstance(item, Term):
                term = item
            else:
                msg = "encountered arg that was not a str or Term object"
                raise ArgError(msg)
            tm_key = term.handle if term.handle else term.value
            prop.value_set.terms[tm_key] = term
            full_term_key = (
                tm_key,
                term.origin_name,
                term.origin_id,
                term.origin_version,
            )
            self.terms[full_term_key] = term

    def rm_node(self, node: Node) -> Node | None:
        """
        Remove a Node from the Model instance.

        Note: A node can't be removed if it is participating in an edge (i.e.,
        if the node is some edge's src or dst attribute).

        Args:
            node: Node to be removed.

        Returns:
            The removed Node, or None if not found.
        """
        if not isinstance(node, Node):
            msg = "arg must be a Node object"
            raise ArgError(msg)
        if not self.contains(node):
            warn(
                f"node '{node.handle}' not contained in model '{self.handle}'",
                stacklevel=2,
            )
            return None
        if self.edges_by_src(node) or self.edges_by_dst(node):
            msg = f"can't remove node '{node.handle}', it is participating in edges"
            raise ValueError(msg)
        for p in node.props:
            with contextlib.suppress(builtins.BaseException):
                del self.props[(node.handle, p.handle)]
        del self.nodes[node.handle]
        self.removed_entities.append(node)
        return node

    def rm_edge(self, edge: Edge) -> Edge | None:
        """
        Remove an Edge instance from the Model instance.

        Args:
            edge: Edge to be removed.

        Returns:
            The removed Edge, or None if not found.
        """
        if not isinstance(edge, Edge):
            msg = "arg must be an Edge object"
            raise ArgError(msg)
        if not self.contains(edge):
            warn(
                f"edge '{edge.triplet}' not contained in model '{self.handle}'",
                stacklevel=2,
            )
            return None
        for p in edge.props:
            with contextlib.suppress(builtins.BaseException):
                k = list(edge.triplet)
                k.append(p.handle)
                del self.props[tuple(k)]
        del self.edges[edge.triplet]
        edge.src = None
        edge.dst = None
        self.removed_entities.append(edge)
        return edge

    def rm_prop(self, prop: Property) -> Property | None:
        """
        Remove a Property instance from the Model instance.

        Args:
            prop: Property to be removed.

        Returns:
            The removed Property, or None if not found.
        """
        if not isinstance(prop, Property):
            msg = "arg must be a Property object"
            raise ArgError(msg)
        if not self.contains(prop):
            warn(
                f"prop '{prop.handle}' not contained in model '{self.handle}'",
                stacklevel=2,
            )
            return
        for okey in prop.belongs:
            owner = prop.belongs[okey]
            (i, att, key) = okey
            del getattr(owner, att)[key]
            k = [owner.handle] if isinstance(owner, Node) else list(owner.triplet)
            k.append(key)
            del self.props[tuple(k)]
        self.removed_entities.append(prop)

    def rm_term(self, term: Term) -> None:
        """Not implemented."""
        if not isinstance(term, Term):
            msg = "arg must be a Term object"
            raise ArgError(msg)

    def assign_edge_end(
        self,
        edge: Edge | None = None,
        end: str | None = None,
        node: Node | None = None,
    ) -> Edge | None:
        """
        Move the src or dst of an Edge to a different Node.

        Note: Both node and edge must be present in the Model instance
        (via add_node and add_edge).

        Args:
            edge: Edge to manipulate.
            end: Edge end to change (src|dst).
            node: Node to be connected.

        Returns:
            The modified Edge, or None if operation failed.
        """
        if not isinstance(edge, Edge):
            msg = "edge= must an Edge object"
            raise ArgError(msg)
        if not isinstance(node, Node):
            msg = "node= must a Node object"
            raise ArgError(msg)
        if end not in ["src", "dst"]:
            msg = "end= must be one of 'src' or 'dst'"
            raise ArgError(msg)
        if not self.contains(edge) or not self.contains(node):
            warn("model must contain both edge and node", stacklevel=2)
            return None
        del self.edges[edge.triplet]
        setattr(edge, end, node)
        self.edges[edge.triplet] = edge
        return edge

    def contains(self, ent: Entity) -> bool | None:
        """
        Ask whether an entity is present in the Model instance.

        Note: Only works on Nodes, Edges, and Properties.

        Args:
            ent: Entity in question.

        Returns:
            True if entity is in model, False otherwise, None if entity type not supported.
        """
        if not isinstance(ent, Entity):
            warn("argument is not an Entity subclass", stacklevel=2)
            return None
        if isinstance(ent, Node):
            return ent in set(self.nodes.values())
        if isinstance(ent, Edge):
            return ent in set(self.edges.values())
        if isinstance(ent, Property):
            return ent in set(self.props.values())
        if isinstance(ent, Term):
            return ent in set(self.terms.values())
        return None

    def edges_in(self, node: Node) -> list[Edge]:
        """
        Get all Edge that have a given Node as their dst attribute.

        Args:
            node: The node.

        Returns:
            List of Edge instances.
        """
        if not isinstance(node, Node):
            msg = "arg must be Node"
            raise ArgError(msg)
        return [self.edges[i] for i in self.edges if i[2] == node.handle]

    def edges_out(self, node: Node) -> list[Edge]:
        """
        Get all Edge that have a given Node as their src attribute.

        Args:
            node: The node.

        Returns:
            List of Edge instances.
        """
        if not isinstance(node, Node):
            msg = "arg must be Node"
            raise ArgError(msg)
        return [self.edges[i] for i in self.edges if i[1] == node.handle]

    def edges_by(self, key: str, item: Node | str) -> list[Edge]:
        """
        Get all Edge that have a given Node as their src or dst or Edge handle as their type attribute.

        Args:
            key: The attribute to search on.
            item: The node or edge handle to search for.

        Returns:
            List of Edge instances.
        """
        if key not in ["src", "dst", "type"]:
            msg = "arg 'key' must be one of src|dst|type"
            raise ArgError(msg)
        if isinstance(item, Node):
            idx = 1 if key == "src" else 2
            return [self.edges[x] for x in self.edges if x[idx] == item.handle]
        return [self.edges[x] for x in self.edges if x[0] == item]

    def edges_by_src(self, node: Node) -> list[Edge]:
        """
        Get all Edge that have a given Node as their src attribute.

        Args:
            node: The node.

        Returns:
            List of Edge instances.
        """
        return self.edges_by("src", node)

    def edges_by_dst(self, node: Node) -> list[Edge]:
        """
        Get all Edge that have a given Node as their dst attribute.

        Args:
            node: The node.

        Returns:
            List of Edge instances.
        """
        return self.edges_by("dst", node)

    def edges_by_type(self, edge_handle: str) -> list[Edge]:
        """
        Get all Edge that have a given edge type (i.e., handle).

        Args:
            edge_handle: The edge type.

        Returns:
            List of Edge instances.
        """
        if not isinstance(edge_handle, str):
            msg = "arg must be str"
            raise ArgError(msg)
        return self.edges_by("type", edge_handle)

    def dget(self, *, refresh: bool = False) -> Model | None:
        """
        Pull model from MDB into this Model instance, based on its handle.

        Note: is a noop if Model.mdb is unset.

        Args:
            refresh: If True, clear cache before retrieving.

        Returns:
            The Model instance, or None if mdb is not set.
        """
        if not self.mdb or self.drv is None:
            return None
        if refresh:
            ObjectMap.clear_cache()
        with self.drv.session() as session:
            result = session.run(
                """
                match p = (s:node {model:$hndl, version:$vers})<-[:has_src]-
                          (r:relationship {model:$hndl, version:$vers})-[:has_dst]->
                          (d:node {model:$hndl, version:$vers})
                return p
                """,
                {"hndl": self.handle, "vers": self.version},
            )
            for rec in result:
                (ns, nr, nd) = rec["p"].nodes
                ns = Node(ns)
                nr = Edge(nr)
                nd = Node(nd)
                ObjectMap.cache[ns.neoid] = ns
                ObjectMap.cache[nr.neoid] = nr
                ObjectMap.cache[nd.neoid] = nd
                nr.src = ns
                nr.dst = nd
                self.nodes[ns.handle] = ns
                self.nodes[nd.handle] = nd
                self.edges[nr.triplet] = nr
            result = session.run(
                """
                match (n:node {model:$hndl, version:$vers})
                where not (n)<--(:relationship)
                return n
                """,
                {"hndl": self.handle, "vers": self.version})
            for rec in result:
                n = Node(rec["n"])
                ObjectMap.cache[n.neoid] = n
                self.nodes[n.handle] = n

        with self.drv.session() as session:
            result = session.run(
                """
                match (n:node {model:$hndl, version:$vers})-[:has_property]->
                (p:property {model:$hndl, version:$vers})
                return id(n), p
                """,
                {"hndl": self.handle, "vers":self.version},
            )
            for rec in result:
                n = ObjectMap.cache.get(rec["id(n)"])
                if n is None:
                    warn(
                        "node with id {nid} not yet retrieved".format(nid=rec["id(n)"]),
                        stacklevel=2,
                    )
                    continue
                p = Property(rec["p"])
                ObjectMap.cache[p.neoid] = p
                self.props[(n.handle, p.handle)] = p
                n.props[p.handle] = p
                p.dirty = -1
        with self.drv.session() as session:
            result = session.run(
                """
                match (r:relationship {model:$hndl, version:$vers})-[:has_property]->
                      (p:property {model:$hndl, version:$vers}) 
                return id(r), p
                """,
                {"hndl": self.handle, "vers":self.version},
            )
            for rec in result:
                e = ObjectMap.cache.get(rec["id(r)"])
                if e is None:
                    warn(
                        "relationship with id {rid} not yet retrieved".format(
                            rid=rec["id(r)"],
                        ),
                        stacklevel=2,
                    )
                    continue
                p = Property(rec["p"])
                ObjectMap.cache[p.neoid] = p
                k = list(e.triplet)
                k.append(p.handle)
                self.props[tuple(k)] = p
                e.props[p.handle] = p
                p.dirty = -1
        return self

    def dput(self) -> None:
        """
        Push this Model's objects to MDB.

        Note: is a noop if Model.mdb is unset.
        """
        if not self.mdb or self.drv is None:
            return
        seen = {}

        def do_(obj: Entity) -> None:
            if id(obj) in seen:
                return
            seen[id(obj)] = 1
            if obj.dirty == 1:
                obj.dput()
            atts = [x for x in type(obj).attspec if type(obj).attspec[x] == "object"]
            for att in atts:
                ent = getattr(obj, att)
                if ent:
                    do_(ent)
            atts = [
                x for x in type(obj).attspec if type(obj).attspec[x] == "collection"
            ]
            for att in atts:
                ents = getattr(obj, att)
                if ents:
                    for ent in ents:
                        do_(ents[ent])

        for e in self.removed_entities:
            # detach
            with self.drv.session() as session:
                session.run(
                    "match (e)-[r]-() where id(e)=$eid delete r return id(e)",
                    {"eid": e.neoid},
                ).consume()
        for e in self.nodes.values():
            do_(e)
        for e in self.edges.values():
            do_(e)
        for e in self.props.values():
            do_(e)
        return



import re

from minicypher.clauses import Match, Return
from minicypher.entities import (
    G,
    N,
    P,
    R,
    _As,
)
from minicypher.functions import (
    And,
    Not,
    Or,
    count,
    exists,
    group,
    labels,
)
from minicypher.statement import Statement

avail_funcs = {x.__name__: x for x in (count, exists, labels, group, And, Or, Not)}


class _engine:
    paths = None

    def __init__(self, use_params=True):
        self.use_params = use_params
        self.error = None
        self.statement = None
        self.params = None
        self.key = ""
        self.path_id = None

    @classmethod
    def set_paths(cls, paths):
        cls.paths = paths

    def parse(self, toks):
        toks = toks
        pth = self.paths
        return self._walk(None, toks, pth)

    def _process_node(self, block):
        ret = None
        if isinstance(block, str):
            if block == "_var":
                ret = N()
            else:
                ret = N(label=block)
        elif isinstance(block, dict):
            if not block["_label"]:
                self.error = {
                    "description": "_node block requires _label key",
                    "block": block,
                }
                return False
            ret = N(label=block["_label"])
            if block.get("_prop"):
                ret._add_props(self._process_prop(block["_prop"]))
            if block.get("_edge") and block.get("_node"):
                n = self._process_node(block["_node"])
                e = self._process_edge(block["_edge"])
                ret = G(ret, e, n)
        else:
            self.error = {
                "description": "Can't process _node block",
                "block": block,
            }
            return False
        return ret

    def _process_prop(self, block, value=None):
        ret = None
        if isinstance(block, str):
            ret = P(handle=block, value=value)
        elif isinstance(block, dict) and block.get("_handle") and block.get("_value"):
            ret = P(handle=block["_handle"], value=block["_value"])
        else:
            self.error = {
                "description": "Can't process _prop block",
                "block": block,
            }
            return False
        return ret

    def _process_edge(self, block):
        ret = None
        if isinstance(block, str):
            ret = R(Type=block)
        elif isinstance(block, dict):
            ret = R(Type=block["_type"])
            if block.get("_dir"):
                ret._dir = block.get("_dir")
            if block.get("_join"):
                # when an edge connects two complex paths
                # define the nodes on the incoming path and new path
                # to link with this edge
                ret._join = block["_join"]
        else:
            self.error = {
                "description": "Can't process _edge block",
                "block": block,
            }
            return False
        return ret

    def _process_func(self, block):
        ret = {}
        if isinstance(block, str):
            as_ = re.split("@", block)
            ret["_func"] = as_[0]
            if len(as_) > 1:
                as_ = as_[1]
            else:
                as_ = None
            if ret["_func"] in avail_funcs:
                # the Func subclass:
                ret["_func"] = avail_funcs[ret["_func"]]
                ret["_func_as"] = as_
            else:
                self.error = {
                    "description": "Sorry, no cypher function '{}' is currently defined".format(
                        ret["_func"],
                    ),
                    "block": block,
                }
                return False  # ERR no such function available
        else:
            self.error = {
                "description": "Block type other than simple string not yet handled for _func",
                "block": block,
            }
            return False
        return ret

    def _create_statement(self, ent, pad):
        match_clause = Match(ent)
        ret_clause = None
        a = []
        if isinstance(pad["_return"], str):
            if pad["_return"] == "_items":
                if isinstance(ent, (N, R)):
                    ret_clause = Return(ent)
                else:
                    ret_clause = Return(*ent.nodes())
            else:
                if (isinstance(ent, N) and ent.label == pad["_return"] and ent.var) or (
                    isinstance(ent, R) and ent.Type == pad["_return"] and ent.var
                ):
                    a.append(ent)
                else:
                    a.extend(
                        [x for x in ent.nodes() if x.label == pad["_return"] and x.var],
                    )
                if not a:
                    self.error = {
                        "description": "No named node to return with label '{}'".format(
                            pad["_return"],
                        ),
                        "ent": ent,
                        "pad": pad,
                    }
                    return False
                ret_clause = Return(*a)
        elif isinstance(pad["_return"], dict):
            retblock = pad["_return"]
            if retblock.get("_path_id"):
                self.path_id = retblock["_path_id"]
            if retblock.get("_nodes"):
                # assume is list-valued
                if not isinstance(retblock["_nodes"], list):
                    self.error = {
                        "description": "_nodes key must point to a list",
                        "ent": ent,
                        "pad": pad,
                    }
                    return False
                labels = {
                    x[0]: x[1]
                    for x in [re.split("@", y) + [None] for y in retblock["_nodes"]]
                }
                if "*" in labels:
                    a.append("*")
                elif isinstance(ent, N) and ent.var:
                    if ent.label in labels:
                        if labels[ent.label]:
                            a.append(_As(ent, labels[ent.label]))
                        else:
                            a.append(ent)
                    if "_var" in labels and ent == pad["_var"]:
                        if labels["_var"]:
                            a.append(_As(ent, labels["_var"]))
                        else:
                            a.append(ent)
                else:
                    for n in ent.nodes():
                        if n.label in labels and n.var:
                            if labels[n.label]:
                                a.append(_As(n, labels[n.label]))
                            else:
                                a.append(n)
                        if "_var" in labels and n == pad["_var"]:
                            if labels["_var"]:
                                a.append(_As(n, labels["_var"]))
                            else:
                                a.append(n)

            if retblock.get("_edges"):
                if not isinstance(retblock["_edges"], list):
                    self.error = {
                        "description": "_edges key must point to a list",
                        "ent": ent,
                        "pad": pad,
                    }
                    return False
                types = {
                    x[0]: x[1]
                    for x in [re.split("@", y) + [None] for y in retblock["_edges"]]
                }
                if isinstance(ent, R) and ent.var:
                    if ent.Type in types:
                        if types[ent.Type]:
                            _As(ent, types[ent.Type])
                        else:
                            a.append(ent)
                    if "_var" in types and ent == pad["_var"]:
                        if types["_var"]:
                            a.append(_As(ent, types["_var"]))
                        else:
                            a.append(ent)
                else:
                    for e in ent.edges():
                        if e.Type in types and e.var:
                            if types[e.Type]:
                                a.append(_As(e, types[e.Type]))
                            else:
                                a.append(e)
            if not a:
                self.error = {
                    "description": "No named nodes or edges matching the path _return spec",
                    "ent": ent,
                    "pad": pad,
                }
                return False
            if retblock.get("_func"):
                f = self._process_func(retblock["_func"])
                if not f:
                    return False  # bad _func spec
                a = [f["_func"](x) for x in a]
                if f.get("_func_as"):
                    a = [_As(x, f["_func_as"]) for x in a]
            ret_clause = Return(*a)
        else:
            self.error = {
                "description": "_return specification not str or dict",
                "ent": ent,
                "pad": pad,
            }
            return False
        self.statement = Statement(match_clause, ret_clause, use_params=self.use_params)
        self.params = self.statement.params
        return True

    def _walk(self, ent, toks, pth):
        if not toks or not pth:
            self.error = {
                "description": "_walk: Either toks or pth is empty",
            }
            return False
        tok = toks[0]
        pad = {}
        parm = None
        if tok in pth:
            # plain token - shouldn't start with _
            self.key = "/".join([self.key, tok]) if self.key else tok
            pth = pth[tok]
        elif any([x.startswith("$") for x in pth]):
            # parameter
            parm = [x for x in pth if x.startswith("$")][0]
            pth = pth[parm]
            # load pad
            pad["_prop"] = P(handle=parm[1:], value=tok)
            # note that the cache key will be interpreted as a regexp
            self.key = (
                "/".join([self.key, "([a-zA-Z0-9_]+)"])
                if self.key
                else "([a-zA-Z0-9_]+)"
            )
        else:
            self.error = {
                "description": f"Token '{tok}' not on valid path",
                "token": tok,
            }
            return False
        # collect/create items req by block in the pad, and then
        # execute operations on these items in standard order below.
        for opn in [x for x in pth if x.startswith("_")]:
            # operations in block
            if opn == "_node":
                if parm:  # processing a parameter
                    pad["_node"] = pth["_node"]
                else:
                    pad["_node"] = self._process_node(pth["_node"])
                    if pth["_node"] == "_var":
                        pad["_var"] = pad["_node"]
            elif opn == "_edge":
                if parm:  # processing a parameter
                    pad["_edge"] = pth["_edge"]
                else:
                    pad["_edge"] = self._process_edge(pth["_edge"])
                    if pth["_edge"] == "_var":
                        pad["_var"] == pad["_edge"]
            elif opn == "_prop":
                # WARN or ERR if pad.get('_prop') is True
                # - a parm was already handled in that case
                pad["_prop"] = self._process_prop(pth["_prop"], value=tok)
            elif opn == "_return":
                if len(toks) == 1:  # we're on the last token
                    # pad['_return'] not empty means we're finished
                    pad["_return"] = pth["_return"]
                else:  # more toks to go...
                    pass  # so ignore it
            elif opn == "_func":
                _func = self._process_func(pth["_func"])
                if not _func:
                    return False  # bad _func spec
                for k in _func:
                    pad[k] = _func[k]

        # pad ready for operations
        new_ent = None
        if pad.get("_prop"):  # add props to incoming entity
            if isinstance(ent, (N, R)):
                ent._add_props(pad["_prop"])
            elif not pad.get("_node") and not pad.get("_edge"):
                self.error = {
                    "description": "Both _edge and _node must be defined here",
                    "token": tok,
                }
                return False  # ERR, if incoming is path, need to specify the node which gets the property (by label)
            else:
                if pad.get("_node"):
                    n = [x for x in ent.nodes() if x.label == pad["_node"]]
                    if not n:
                        self.error = {
                            "description": "Node specified by _node is not present",
                            "token": tok,
                        }
                        return False  # specified node can't be found in ent
                    n[0]._add_props(pad["_prop"])
                    pad["_node"] = None
                    pad["_prop"] = None
                if pad.get("_edge"):
                    e = [x for x in ent.edges() if x.Type == pad["_edge"]]
                    if not e:
                        self.error = {
                            "description": "Edge specified by _edge is not present",
                            "token": tok,
                        }
                        return False  # specified edge not found in ent
                    e[0]._add_props(pad["_prop"])
                    pad["_edge"] = None
                    pad["_prop"] = None

        if pad.get("_node"):  # new entity
            new_ent = pad["_node"]
        if pad.get("_edge"):
            if not ent:  # no input entity
                self.error = {
                    "description": "No incoming entity to apply _edge to here",
                    "token": tok,
                }
                return False  # ERR
            if not new_ent:  # no new entity
                self.error = {
                    "description": "No new entity to link to",
                    "token": tok,
                }
                return False  # different ERR
            if isinstance(ent, N) and isinstance(new_ent, N):
                # simple relationship
                new_ent = pad["_edge"].relate(ent, new_ent)
            else:
                new_ent = G(ent, pad["_edge"], new_ent)
        if pad.get("_return"):  # we made it
            if self._create_statement(new_ent or ent, pad):
                return True
            return False
        if len(toks) == 1:
            self.error = {
                "description": "Reached end of path, but found no _return spec",
                "token": tok,
            }
            return False  # ERR reached end of toks, no _return found
        return self._walk(new_ent or ent, toks[1:], pth)



"""
mdb.loaders: load models into an MDB instance consistently
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from minicypher.clauses import (
    Create,
    Match,
    Merge,
    Remove,
)
from minicypher.entities import N, R, _plain_var
from minicypher.statement import Statement
from tqdm import tqdm

from bento_meta.mdb.writeable import WriteableMDB

if TYPE_CHECKING:
    from minicypher.entities import Entity as CypherEntity

    from bento_meta.entity import Entity
    from bento_meta.model import Model


class MDFProtocol(Protocol):
    """
    Protocol defining the expected interface for MDF objects from bento-mdf package.

    This protocol is used to provide type hints without creating a circular dependency,
    since bento-mdf depends on bento-meta.
    """

    model: Model


def load_mdf(mdf: MDFProtocol, mdb: WriteableMDB, _commit: str | None = None) -> None:
    """Load an MDF object into an MDB instance."""
    load_model(mdf.model, mdb, _commit)


def load_model(model: Model, mdb: WriteableMDB, _commit: str | None = None) -> None:
    """Load a model object into an MDB instance."""
    if not isinstance(mdb, WriteableMDB):
        msg = "mdb object must be a WriteableMDB"
        raise TypeError(msg)
    cstmts = load_model_statements(model, _commit)
    for stmt in tqdm(cstmts):
        mdb.put_with_statement(str(stmt), stmt.params)


def load_model_statements(model: Model, _commit: str | None = None) -> list[Statement]:
    """
    Create Cypher statements from a model to load it de novo into an MDB instance.

    :param :class:`mdb.Model` model: Model instance for loading
    :param str _commit: 'Commit string' for marking entities in DB. If set,
        this will override _commit attributes already existing on Model entities.
    """
    c_statements: list[Statement] = []
    c_nodes = {}
    for nd in model.nodes:
        node = model.nodes[nd]
        c_node = _c_entity(node, model, _commit)
        c_statements.append(
            Statement(
                Merge(c_node),
                use_params=True,
            ),
        )
        c_nodes[node.handle] = c_node
        c_props = []
        if node.tags:
            c_statements.extend(
                _tag_statements(node, c_node, _commit),
            )
        if node.props:
            c_statements.extend(
                _prop_statements(node, c_node, model, _commit),
            )
        if node.concept:
            c_statements.extend(
                _annotate_statements(node, c_node, _commit),
            )
    # node nodes and node-property nodes now exist
    # nodes are linked to properties
    for rl in model.edges:
        edge = model.edges[rl]
        c_edge = _c_entity(edge, model, _commit)
        c_src = _c_entity(edge.src, model, _commit)
        c_dst = _c_entity(edge.dst, model, _commit)

        if edge.multiplicity:
            c_edge._add_props({"multiplicity": edge.multiplicity})
        if edge.is_required:
            c_edge._add_props({"is_required": edge.is_required})
        # ensure uniqueness for merge
        c_edge._add_props({"__u": str(rl)})
        c_statements.extend(
            [
                Statement(
                    Create(c_edge),
                    use_params=True,
                ),
                Statement(
                    Match(c_edge, c_src, c_dst),
                    Merge(
                        R(Type="has_src").relate(_plain_var(c_edge), _plain_var(c_src)),
                    ),
                    Merge(
                        R(Type="has_dst").relate(_plain_var(c_edge), _plain_var(c_dst)),
                    ),
                    use_params=True,
                ),
            ],
        )
        if edge.tags:
            c_statements.extend(
                _tag_statements(edge, c_edge, _commit),
            )
        if edge.props:
            c_statements.extend(
                _prop_statements(edge, c_edge, model, _commit),
            )
        if edge.concept:
            c_statements.extend(
                _annotate_statements(edge, c_edge, _commit),
            )
        c_statements.append(
            Statement(
                Match(c_edge),
                Remove(c_edge, prop="__u"),
                use_params=True,
            ),
        )
    # edge node and edge-property nodes now exist

    # now go through all properties that the model object knows about
    # - these should already have been created or merged, but
    # - if the property list on the model is missing any properties
    # - that nodes or edges have on themselves, then this indicates
    # - a bug/inconsistency  - and the property won't receive its
    # - value_set/term list in the DB in the following code.

    for pr in [x for x in model.props.values() if x.value_domain == "value_set"]:
        c_value_set = _c_entity(pr.value_set, model, _commit)
        c_prop = _c_entity(pr, model, _commit)
        c_statements.extend(
            [
                Statement(
                    Merge(c_value_set),
                    use_params=True,
                ),
                Statement(
                    Match(c_prop, c_value_set),
                    Merge(
                        R(Type="has_value_set").relate(
                            _plain_var(c_prop),
                            _plain_var(c_value_set),
                        ),
                    ),
                    use_params=True,
                ),
            ],
        )
        if pr.concept:
            c_statements.extend(
                _annotate_statements(pr, c_prop, _commit),
            )
        for tm in pr.terms.values():
            c_term = _c_entity(tm, model, _commit)
            c_statements.extend(
                [
                    Statement(
                        Merge(c_term),
                        use_params=True,
                    ),
                    Statement(
                        Match(c_value_set, c_term),
                        Merge(
                            R(Type="has_term").relate(
                                _plain_var(c_value_set),
                                _plain_var(c_term),
                            ),
                        ),
                        use_params=True,
                    ),
                ],
            )
    return c_statements


def _c_entity(ent: Entity, model: Model | None, _commit: str | None = None) -> N:
    label = type(ent).__name__.lower()
    c_ent = None

    # translate labels
    if label == "edge":
        label = "relationship"
    if label == "valueset":
        label = "value_set"

    # special handling
    if label == "term":
        c_ent = N(label=label, props={"value": ent.value})
        if ent.origin_name:
            c_ent._add_props({"origin_name": ent.origin_name})
        if ent.origin_id:
            c_ent._add_props({"origin_id": ent.origin_id})
        if ent.origin_version:
            c_ent._add_props({"origin_version": ent.origin_version})
        if ent.origin_definition:
            c_ent._add_props({"origin_defintion": ent.origin_definition})
        if ent.handle:
            c_ent._add_props({"handle": ent.handle})
    elif label == "value_set":
        c_ent = N(label="value_set", props={"handle": ent.handle})
        if ent.url:
            c_ent._add_props({"url": ent.url})
    elif label == "concept":
        c_ent = N(label="concept")
    elif label == "tag":
        c_ent = N(label="tag", props={"key": ent.key, "value": ent.value})

    else:
        model_handle = model.handle if model else None
        c_ent = N(label=label, props={"handle": ent.handle, "model": model_handle})
    # all ents
    if _commit:
        c_ent._add_props({"_commit": _commit})
    elif ent._commit:
        c_ent._add_props({"_commit": ent._commit})
    if ent.nanoid:
        c_ent._add_props({"nanoid": ent.nanoid})
    if ent.desc:
        c_ent._add_props({"desc": ent.desc})
    return c_ent


def _tag_statements(
    ent: Entity,
    c_ent: CypherEntity,
    _commit: str | None = None,
) -> list[Statement]:
    if not ent.tags:
        return []

    c_tags = [_c_entity(t, None, _commit) for t in ent.tags.values()]
    return [
        Statement(
            Match(c_ent),
            Merge(R(Type="has_tag").relate(_plain_var(c_ent), ct)),
            use_params=True,
        )
        for ct in c_tags
    ]


def _prop_statements(
    ent: Entity,
    c_ent: CypherEntity,
    model: Model | None,
    _commit: str | None = None,
) -> list[Statement]:
    stmts = []
    for p in ent.props.values():
        c_prop = _c_entity(p, model, _commit)
        c_prop._add_props({"value_domain": p.value_domain})
        stmts.extend(
            [
                Statement(
                    Merge(c_prop),
                    use_params=True,
                ),
                Statement(
                    Match(c_ent, c_prop),
                    Merge(
                        R(Type="has_property").relate(
                            _plain_var(c_ent),
                            _plain_var(c_prop),
                        ),
                    ),
                    use_params=True,
                ),
            ],
        )
        if p.tags:
            stmts.extend(
                _tag_statements(p, c_prop, _commit),
            )
    return stmts


def _annotate_statements(
    ent: Entity,
    c_ent: CypherEntity,
    _commit: str | None = None,
) -> list[Statement]:
    stmts = []
    if not ent.concept:
        return []
    c_concept = _c_entity(ent.concept, None, _commit)
    stmts = []
    for tm in ent.concept.terms.values():
        c_term = _c_entity(tm, None, _commit)
        stmts.extend(
            [
                Statement(
                    Merge(c_term),
                    use_params=True,
                ),
                Statement(
                    Match(c_ent),
                    Merge(R(Type="has_concept").relate(_plain_var(c_ent), c_concept)),
                    use_params=True,
                ),
                Statement(
                    Match(R(Type="has_concept").relate(c_ent, c_concept), c_term),
                    Merge(
                        R(Type="represents").relate(
                            _plain_var(c_term),
                            _plain_var(c_concept),
                        ),
                    ),
                    use_params=True,
                ),
            ],
        )
    return stmts



"""
bento_meta.entity
=================

This module contains
* `Entity`, the base class for metamodel objects,
* the `CollValue` class to manage collection-valued attributes, and
* the `ArgError` exception.
"""

from __future__ import annotations

from collections import UserDict
from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING:
    import neo4j

    from bento_meta.object_map import ObjectMap


class ArgError(Exception):
    """Exception for method argument errors."""


class Entity:
    """
    Base class for all metamodel objects.

    Entity contains all the magic for metamodel objects such as
    `bento_meta.objects.Node` and 'bento_meta.object.Edge`. It will rarely
    be used directly. Entity redefines `__setattr__` and `__getattr__` to
    enable graph database object mapping under the hood.

    The Entity class also defines private and declared attributes that are
    common to all metamodel objects. It provides the machinery to manage
    private attributes separately from declared attributes, and to raise
    exceptions when attempts are made to access attributes that are not
    declared.
    """

    pvt_attr: ClassVar[list[str]] = [
        "pvt",
        "neoid",
        "dirty",
        "removed_entities",
        "attspec",
        "mapspec",
        "belongs",
    ]
    defaults = ({},)
    attspec_: ClassVar[dict[str, str]] = {
        "_id": "simple",
        "nanoid": "simple",
        "desc": "simple",
        "_next": "object",
        "_prev": "object",
        "_from": "simple",
        "_to": "simple",
        "_commit": "simple",
        "tags": "collection",
    }
    attspec = attspec_
    mapspec_: ClassVar[
        dict[str, str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None]
    ] = {
        "label": None,
        "key": "_id",
        "property": {
            "_id": "id",
            "desc": "desc",
            "_from": "_from",
            "_to": "_to",
            "_commit": "_commit",
            "nanoid": "nanoid",
        },
        "relationship": {
            "_next": {"rel": ":_next>", "end_cls": set()},
            "_prev": {"rel": ":_prev>", "end_cls": set()},
            "tags": {"rel": ":has_tag", "end_cls": {"Tag"}},
        },
    }
    object_map: ClassVar[ObjectMap | None] = None

    def __init__(self, init: dict | neo4j.graph.Node | Entity | None = None) -> None:
        """
        Entity constructor. Always called by subclasses.

        Args:
            init: One of the following:
                - dict: A dict of attribute names and values. Undeclared attributes
                    are ignored.
                - neo4j.graph.Node: A Neo4j node object to be stored as a model object.
                - Entity: An Entity (of matching subclass) used to duplicate another
                    model object.
        """
        if not set(type(self).attspec.values()) <= {"simple", "object", "collection"}:
            msg = "unknown attribute type in attspec"
            raise ArgError(msg)

        # private
        self.pvt = {}
        self.neoid = None
        self.dirty = 1
        self.removed_entities = []
        self.belongs = {}
        # merge to universal map - no, do in the subclasses
        # type(self).mergespec()

        if init:
            if isinstance(init, Entity):
                self.set_with_entity(init)
            elif isinstance(init, dict):
                self.set_with_dict(init)
            elif (
                type(init).__name__ == "Node"
            ):  # neo4j.graph.Node - but don't want to import that
                self.set_with_node(init)
        for att in type(self).attspec:
            if att not in self.__dict__:
                if self.attspec[att] == "collection":
                    setattr(self, att, CollValue({}, owner=self, owner_key=att))
                else:
                    setattr(self, att, None)

    @classmethod
    def mapspec(cls) -> dict[str, str | dict[str, str]]:
        """
        Get object to database mapping specification.

        Is a class method, not a property.
        """
        if not hasattr(cls, "_mapspec"):
            cls.mergespec()
        return cls._mapspec

    @classmethod
    def default(cls, propname: str) -> Any:
        """
        Return a default value for the property named.

        Args:
            propname: Name of the property to get default for.

        Returns:
            Default value if defined, None otherwise.
        """
        if cls.defaults.get(propname):
            return cls.defaults[propname]
        return None

    # @classmethod
    def get_by_id(self, id: str) -> Entity | None:
        """
        Get an object from the db with the id attribute (not the Neo4j id).

        Args:
            id: Value of id for desired object.

        Returns:
            A new object if found, None otherwise.
        """
        if self.object_map:
            print(f"  > now in entity.get_by_id where self is {self}")
            print(f"  > and class is {self.__class__}")
            return self.object_map.get_by_id(self, id)
        print("    _NO_ cls.object_map detected")
        return None

    @property
    def dirty(self) -> int:
        """
        Flag whether this instance has been changed since retrieval from the database.

        Set to -1, ensure that the next time an attribute is accessed, the instance
        will retrieve itself from the database.
        """
        return self.pvt["dirty"]

    @dirty.setter
    def dirty(self, value: int) -> None:
        """Set dirty flag."""
        self.pvt["dirty"] = value

    @property
    def removed_entities(self) -> list[Any]:
        """Return list of removed entities."""
        return self.pvt["removed_entities"]

    @removed_entities.setter
    def removed_entities(self, value: list[Any]) -> None:
        """Set list of removed entities."""
        self.pvt["removed_entities"] = value

    @property
    def object_map(self) -> ObjectMap | None:
        """Return object map."""
        return self.pvt.get("object_map")

    @object_map.setter
    def object_map(self, value: ObjectMap | None) -> None:
        """Set object map."""
        self.pvt["object_map"] = value

    @property
    def belongs(self) -> dict[tuple[int, str, str] | tuple[int, str], Entity]:
        """Return dict that stores information on the owners (referents) of this instance in the model."""
        return self.pvt["belongs"]

    @belongs.setter
    def belongs(
        self,
        value: dict[tuple[int, str, str] | tuple[int, str], Entity],
    ) -> None:
        """Set belongs dict."""
        self.pvt["belongs"] = value

    def clear_removed_entities(self) -> None:
        """Clear the list of removed entities."""
        self.pvt["removed_entities"] = []

    def set_with_dict(self, init: dict) -> None:
        """Set the entity with a dict."""
        for att in type(self).attspec:
            if att in init:
                if type(self).attspec[att] == "collection":
                    setattr(self, att, CollValue(init[att], owner=self, owner_key=att))
                else:
                    setattr(self, att, init[att])

    def set_with_node(self, init: neo4j.graph.Node) -> None:
        """Set the entity with a Neo4j node."""
        # this unsets any attribute that is not present in the Node's properties
        for att in [a for a in type(self).attspec if type(self).attspec[a] == "simple"]:
            patt = type(self).mapspec()["property"][att]
            if patt in init:
                setattr(self, att, init[patt])
            else:
                setattr(self, att, None)
        self.neoid = init.id

    def set_with_entity(self, ent: Entity) -> Entity:
        """Set the entity with another entity."""
        if not isinstance(self, type(ent)):
            msg = (
                f"class mismatch: I am a {type(self).__name__}, "
                f"but arg is a {type(ent).__name__}"
            )
            raise ArgError(msg)
        for k in type(self).attspec:
            atts = type(self).attspec[k]
            if atts in {"simple", "object"}:
                setattr(self, k, getattr(ent, k))
            elif atts == "collection":
                setattr(self, k, CollValue(getattr(ent, k), owner=self, owner_key=k))
            else:
                msg = f"unknown attribute type '{atts}'"
                raise RuntimeError(msg)
        for okey in ent.belongs:
            self.belongs[okey] = ent.belongs[okey]
        self.neoid = ent.neoid
        self.dirty = 1
        return self

    def __getattribute__(self, name: str) -> Any:  # noqa: ANN401
        """Get the attribute of the entity."""
        if name in type(self).attspec:
            # declared attr, send to __getattr__ for magic
            return self.__getattr__(name)
        return object.__getattribute__(self, name)

    def __getattr__(self, name: str) -> Any:  # noqa: ANN401
        """Get the attribute of the entity."""
        if name in type(self).pvt_attr:
            return self.__dict__["pvt"].get(name)
        if name in type(self).attspec:
            if name not in self.__dict__ or self.__dict__[name] is None:
                return None
            if type(self).attspec[name] == "object" and self.__dict__[name].dirty < 0:
                # magic - lazy getting
                self.__dict__[name].dget()
            return self.__dict__[name]
        msg = (
            f"get: attribute '{name}' neither private nor declared "
            f"for subclass {type(self).__name__}"
        )
        raise AttributeError(msg)

    def __setattr__(self, name: str, value: Any) -> None:  # noqa: ANN401
        """Set the attribute of the entity."""
        if name == "pvt":
            self.__dict__["pvt"] = value
        elif name in type(self).pvt_attr:
            self.__dict__["pvt"][name] = value
        elif name in type(self).attspec:
            self._check_value(name, value)
            self._set_declared_attr(name, value)
        else:
            msg = (
                f"get: attribute '{name}' neither private nor declared "
                f"for subclass {type(self).__name__}"
            )
            raise AttributeError(msg)

    def _set_declared_attr(self, name: str, value: Any) -> None:
        """Set the declared attribute of the entity."""
        atts = type(self).attspec[name]
        if atts == "simple":
            pass
        elif atts == "object":
            oldval = self.__dict__.get(name)
            if oldval and oldval == value:
                return  # a wash
            if isinstance(value, Entity):
                value.belongs[(id(self), name)] = self
        elif atts == "collection":
            if isinstance(value, dict):
                value = CollValue(value, owner=self, owner_key=name)
            if isinstance(value, list):  # convert list of objs to CollValue
                d = {}
                for v in value:
                    d[getattr(v, type(v).mapspec()["key"])] = v
                value = CollValue(d, owner=self, owner_key=name)
        else:
            msg = f"unknown attspec value '{atts}'"
            raise RuntimeError(msg)
        self.dirty = 1
        self.__dict__[name] = value

    def __delattr__(self, name: str) -> None:
        """Delete the attribute of the entity."""
        del self.__dict__[name]

    def _check_init(self, init: dict) -> None:
        """Check the initial value of the entity."""
        for att in type(self).attspec:
            if init[att]:
                self._check_value(att, init[att])

    def _check_value(self, att: str, value: Any) -> None:
        """Check the value of the attribute."""
        spec = type(self).attspec[att]
        if spec == "simple":
            if not (isinstance(value, (bool, float, int, str))) and value is not None:
                msg = f"value for key '{att}' is not a simple scalar"
                raise ArgError(msg)
        elif spec == "object":
            if not (isinstance(value, Entity) or value is None):
                msg = f"value for key '{att}' is not an Entity subclass"
                raise ArgError(msg)
        elif spec == "collection":
            if not (isinstance(value, (dict, list, CollValue))):
                msg = f"value for key '{att}' is not a dict,list, or CollValue"
                raise AttributeError(msg)
        else:
            msg = f"unknown attribute type '{spec}' for attribute '{att}' in attspec"
            raise ArgError(msg)

    def dup(self) -> Entity:
        """Duplicate the object, but not too deeply."""
        return type(self)(self)

    def delete(self) -> None:
        """Delete self from the database."""
        # unlink from other entities
        for okey in self.belongs:
            owner = self.belongs[okey]
            (oid, *att) = okey
            if len(att) == 2:
                del getattr(owner, att[0])[att[1]]
            else:
                setattr(owner, att[0], None)

    def dget(self, *, refresh: bool = False) -> Entity | None:
        """
        Update self from the database.

        Args:
            refresh: If True, force a retrieval from db. If False, retrieve from cache
                and don't disrupt changes already made.

        Returns:
            The entity if found, None otherwise.
        """
        if type(self).object_map:
            return type(self).object_map.get(self, refresh=refresh)
        return None

    def dput(self) -> None:
        """
        Put self to the database.

        This will set the neoid property if not yet set.
        """
        if type(self).object_map:
            return type(self).object_map.put(self)
        return None

    def rm(self, *, force: bool = False) -> None:
        """
        Delete self from the database.

        The object instance survives.

        Args:
            force: If True, detach and delete the node.
        """
        if type(self).object_map:
            return type(self).object_map.rm(self, force=force)
        return None

    @classmethod
    def attr_doc(cls) -> str:
        """Create a docstring for declared attributes on class as configured."""

        def str_for_obj(thing: set | str) -> str:
            """Convert a set of classes to a string."""
            if isinstance(thing, set):
                return "|".join(thing)
            return thing

        (first, *rest) = cls.__doc__.split("\n")
        if cls.__name__ == "Entity":
            first += " Posesses the following attributes:"
        else:
            first += " Posesses all :class:`Entity` attributes, plus the following:"
        doc = f"""\
.. py:class:: {cls.__name__}

{first}

"""
        for att in [x for x in cls.attspec_ if cls.attspec[x] == "simple"]:
            doc += """\
  .. py:attribute:: {att}
       :type: simple
""".format(
                att=cls.__name__.lower() + "." + att,
            )
        for att in [x for x in cls.attspec_ if cls.attspec[x] == "object"]:
            doc += """\
  .. py:attribute:: {att}
       :type: {obj}
""".format(
                att=cls.__name__.lower() + "." + att,
                obj=str_for_obj(cls.mapspec_["relationship"][att]["end_cls"]),
            )
        for att in [x for x in cls.attspec_ if cls.attspec[x] == "collection"]:
            doc += """\
  .. py:attribute:: {att}
       :type: collection of {obj}
""".format(
                att=cls.__name__.lower() + "." + att,
                obj=str_for_obj(cls.mapspec_["relationship"][att]["end_cls"]),
            )
        doc += "\n"
        return doc

    def get_label(
        self,
    ) -> str | dict[str, str] | dict[str, dict[str, str | set[str]]] | None:
        """Return type of entity as label."""
        return self.mapspec_["label"]

    def get_attr_dict(self) -> dict[str, str]:
        """
        Return simple attributes set for Entity as a dict.

        Attr values are converted to strings. Doesn't include attrs with None values.

        Returns:
            Dictionary of attribute names to string values.
        """
        return {
            k: str(getattr(self, k))
            for k in self.attspec
            if self.attspec[k] == "simple" and getattr(self, k) is not None
        }


class CollValue(UserDict):
    """
    A UserDict for housing Entity collection attributes.

    This class contains a hook for recording the Entity that
    owns the value that is being set. The value is marked as belonging
    to the containing object, not this collection object.
    It also protects against adding arbitrarily typed elements to the
    collection; it throws unless a value to set is an Entity.

    Attributes:
        owner: Entity object of which this collection is an attribute.
        owner_key: The attribute name of this collection on the owner.
    """

    def __init__(
        self,
        init: dict | None = None,
        *,
        owner: Entity,
        owner_key: str,
    ) -> None:
        """
        Initialize the CollValue.

        Args:
            init: The initial value for the collection.
            owner: The entity instance of which this collection is an attribute.
            owner_key: The attribute name of this collection on the owner.
        """
        self.__dict__["__owner"] = owner
        self.__dict__["__owner_key"] = owner_key
        super().__init__(init)

    @property
    def owner(self) -> Entity:
        """The entity instance of which this collection is an attribute."""
        return self.__dict__["__owner"]

    @property
    def owner_key(self) -> str:
        """The attribute name of this collection on the `owner`."""
        return self.__dict__["__owner_key"]

    def __setitem__(self, name: str, value: Entity) -> None:
        """Set the value for the collection."""
        if not isinstance(value, Entity):
            msg = (
                "a collection-valued attribute can only accept Entity members, "
                "not '{type(value)}'s"
            )
            raise ArgError(msg)
        if name in self:
            oldval = self.data.get(name)
            if oldval and oldval == value:
                return  # a wash
        value.belongs[(id(self.owner), self.owner_key, name)] = self.owner
        # smudge the owner
        self.owner.dirty = 1
        self.data[name] = value
        return

    def __getitem__(self, name: str) -> Any:  # noqa: ANN401
        """Get the value for the collection."""
        if name not in self.data:
            return None
        if self.data[name].dirty < 0:
            self.data[name].dget()
        return self.data[name]

    def __delitem__(self, name: str) -> None:
        """Delete the value for the collection."""
        if name in self.data: # cleanup belongs and set owner dirty
            entity = self.data[name]
            belongs_key = (id(self.owner), self.owner_key, name)
            if belongs_key in entity.belongs:
                del entity.belongs[belongs_key]
            self.owner.dirty = 1
        super().__delitem__(name)



"""
ToolsMDB: subclass of 'WriteableMDB' to support interactions with the MDB.

EntityValidator: validates that entities have required attributes.
"""

from __future__ import annotations

import csv
import logging
from collections.abc import Iterable
from importlib.util import find_spec
from logging.config import fileConfig
from pathlib import Path
from subprocess import check_call
from sys import executable
from typing import TYPE_CHECKING, ClassVar

from minicypher.clauses import (
    As,
    Collect,
    DetachDelete,
    Match,
    Merge,
    OptionalMatch,
    Return,
    With,
)
from minicypher.entities import N0, R0, G, N, P, R, T, _plain_var
from minicypher.functions import count
from minicypher.statement import Statement

from bento_meta.entity import Entity
from bento_meta.mdb import make_nanoid, read_txn_data, read_txn_value
from bento_meta.mdb.writeable import WriteableMDB, write_txn
from bento_meta.objects import (
    Concept,
    Edge,
    Node,
    Predicate,
    Property,
    Tag,
    Term,
    ValueSet,
)
from bento_meta.tf_objects import (
    Transform,
    TfStep,
)
from minicypher.statement import Statement
from minicypher.clauses import (
    As,
    Collect,
    Match,
    Merge,
    DetachDelete,
    OptionalMatch,
    Return,
    With,
)

from minicypher.entities import N0, R0, G, N, P, R, T, _plain_var
from minicypher.functions import count
if TYPE_CHECKING:
    from neo4j import Record

# logging stuff
log_ini_path = Path(__file__).parents[2].joinpath("logs/log.ini")
log_file_path = Path(__file__).parents[2].joinpath(f"logs/{__name__}.log")
fileConfig(log_ini_path, defaults={"logfilename": log_file_path.as_posix()})
logger = logging.getLogger(__name__)


class ToolsMDB(WriteableMDB):
    """:class:`bento_meta.mdb.writeable.WriteableMDB` subclass with mdb-tools."""

    def __init__(self, uri: str | None, user: str | None, password: str | None) -> None:
        """Initialize a :class:`ToolsMDB` object."""
        super().__init__(uri=uri, user=user, password=password)

    class EntityNotUniqueError(Exception):
        """Entity's attributes identify more than 1 property graph node in an MDB."""

    class EntityNotFoundError(Exception):
        """Entity's attributes fail to identify a property graph node in an MDB."""

    class PatternNotUniqueError(Exception):
        """Match pattern's attributes identify more than 1 triple in an MDB."""

    class PatternNotFoundError(Exception):
        """Match pattern's attributes fail to identify a triple in an MDB."""

    @read_txn_value  # type: ignore[reportArgumentType]
    def _get_entity_count(self, entity: Entity) -> list[Record]:
        """
        Return count of given entity found in an MDB.

        - If count = 0, entity with given attributes not found in MDB.
        - If count = 1, entity with given attributes is unique in MDB.
        - If count > 1, more attributes needed to uniquely id entity in MDB.

        Args:
            entity: The entity to count.

        Returns:
            List with count value.
        """
        ent = N(label=entity.get_label(), props=entity.get_attr_dict())

        stmt = Statement(
            Match(ent),
            Return(count(ent)),
            As("entity_count"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "entity_count")  # type: ignore[reportReturnType]

    @read_txn_value  # type: ignore[reportArgumentType]
    def _get_pattern_count(self, pattern: T | G) -> list[Record]:
        """
        Return count of given match pattern (triple or path) found in an MDB.

        Property graph triple: (n)-[r]->(m)
        Path: set of overlapping triples

        - If count = 0, pattern with given attributes not found in MDB.
        - If count = 1, pattern with given attributes is unique in MDB.
        - If count > 1, more attributes needed to uniquely id pattern in MDB.

        Args:
            pattern: The pattern (triple or path) to count.

        Returns:
            List with count value.
        """
        stmt = Statement(
            Match(pattern),
            Return(count("*")),
            As("pattern_count"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "pattern_count")  # type: ignore[reportReturnType]

    def validate_entity_unique(self, entity: Entity) -> None:
        """
        Validate that the given entity occurs once (& only once) in an MDB.

        Note: doesn't validate the entity itself because not all entity attributes
        are necessarily required to locate an entity in the MDB.
        (e.g. handle and model OR nanoid alone can identify a node)

        Args:
            entity: The entity to validate.

        Raises:
            EntityNotUniqueError: If entity attributes match multiple property
                graph nodes in the MDB.
            EntityNotFoundError: If entity attributes don't match any in the MDB.
        """
        ent_count = int(self._get_entity_count(entity)[0])
        if ent_count > 1:
            logger.error(str(self.EntityNotUniqueError))
            raise self.EntityNotUniqueError
        if ent_count < 1:
            logger.error(str(self.EntityNotFoundError))
            raise self.EntityNotFoundError

    def validate_entities_unique(self, entities: Iterable[Entity]) -> None:
        """Run self.validate_entity_unique() over multiple entities."""
        for entity in entities:
            self.validate_entity_unique(entity)

    def validate_pattern_unique(self, pattern: T | G) -> None:
        """
        Validate that the given match pattern occurs once (& only once) in an MDB.

        Args:
            pattern: The pattern to validate.

        Raises:
            PatternNotUniqueError: If pattern attributes match multiple triples
                in the MDB.
            PatternNotFoundError: If pattern attributes don't match any in the MDB.
        """
        pattern_count = int(self._get_pattern_count(pattern)[0])
        if pattern_count > 1:
            logger.error(
                str(
                    self.PatternNotUniqueError(
                        f"Pattern: {pattern.pattern()} not unique.",
                    ),
                ),
            )
            msg = f"Pattern: {pattern.pattern()} not unique."
            raise self.PatternNotUniqueError(msg)
        if pattern_count < 1:
            logger.error(
                str(
                    self.PatternNotFoundError(
                        f"Pattern: {pattern.pattern()} not found.",
                    ),
                ),
            )
            msg = f"Pattern: {pattern.pattern()} not found."
            raise self.PatternNotFoundError(msg)

    @write_txn  # type: ignore[reportArgumentType]
    def remove_entity_from_mdb(self, entity: Entity) -> list[Record]:
        """
        Remove given Entity node from the database.

        Accepts the following bento-meta Entities:
            Concept, Node, Predicate, Property, Edge, Term

        Args:
            entity: The entity to remove.

        Returns:
            List of Records from the transaction.
        """
        self.validate_entity_unique(entity)

        ent_label = entity.get_label()
        ent_attrs = entity.get_attr_dict()

        ent = N(label=ent_label, props=ent_attrs)

        stmt = Statement(Match(ent), DetachDelete(ent._var), use_params=True)

        qry = str(stmt)
        parms = stmt.params

        logger.info("Removing %s entity with with properties: %s", ent_label, ent_attrs)
        return (qry, parms)  # type: ignore[reportReturnType]

    @write_txn  # type: ignore[reportArgumentType]
    def add_entity_to_mdb(
        self,
        entity: Entity,
        _commit: str | None = None,
    ) -> list[Record]:
        """
        Add given Entity node to MDB instance.

        Args:
            entity: The entity to add.
            _commit: Optional commit string to tag the entity.

        Returns:
            List of Records from the transaction.
        """
        EntityValidator.validate_entity(entity)
        EntityValidator.validate_entity_has_attribute(entity, "nanoid")

        if _commit:
            entity._commit = _commit

        ent_label = entity.get_label()
        ent_attrs = entity.get_attr_dict()

        ent = N(label=ent_label, props=ent_attrs)  # type: ignore[arg-type]

        if isinstance(entity, Edge):
            src = N(label="node", props=entity.src.get_attr_dict())
            dst = N(label="node", props=entity.dst.get_attr_dict())
            src_rel = R(Type="has_src")
            dst_rel = R(Type="has_dst")
            src_trip = src_rel.relate(_plain_var(ent), _plain_var(src))
            dst_trip = dst_rel.relate(_plain_var(ent), _plain_var(dst))
            stmt = Statement(
                Merge(ent),
                Merge(src),
                Merge(dst),
                Merge(src_trip),
                Merge(dst_trip),
                use_params=True,
            )
        else:
            stmt = Statement(Merge(ent), use_params=True)

        qry = str(stmt)
        parms = stmt.params

        logger.info(
            "Merging new %s node with properties: %s into the MDB",
            ent_label,
            ent_attrs,
        )

        return (qry, parms)  # type: ignore[reportReturnType]

    @read_txn_value  # type: ignore[reportArgumentType]
    def get_concept_nanoids_linked_to_entity(
        self,
        entity: Entity,
        mapping_source: str | None = None,
    ) -> list[Record]:
        """
        Return list of concept nanoids linked to given entity.

        The concept(s) have "represents" or "has_concept" relationships tagged with
        the given mapping source. If no mapping source is provided, return all.

        Args:
            entity: The entity to get concepts for.
            mapping_source: Optional mapping source to filter by.

        Returns:
            List of concept nanoids.
        """
        self.validate_entity_unique(entity)

        ent = N(label=entity.get_label(), props=entity.get_attr_dict())
        concept = N(label="concept")
        if isinstance(entity, Term):
            rel = R(Type="represents")
        else:
            rel = R(Type="has_concept")
        ent_trip = rel.relate(ent, concept)

        if mapping_source is not None:
            tag = N(
                label="tag",
                props={"key": "mapping_source", "value": mapping_source},
            )
            tag_trip = T(concept, R(Type="has_tag"), tag)
            path = G(ent_trip, tag_trip)
        else:
            path = ent_trip

        stmt = Statement(
            Match(path),
            Return(f"{concept._var}.nanoid"),
            As("concept_nanoids"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params
        logger.debug("%s; %s", qry, parms)

        return (qry, parms, "concept_nanoids")  # type: ignore[reportReturnType]

    @write_txn  # type: ignore[reportArgumentType]
    def add_relationship_to_mdb(
        self,
        relationship_type: str,
        src_entity: Entity,
        dst_entity: Entity,
        _commit: str = "",
    ) -> list[Record]:
        """
        Add relationship between given entities in MDB.

        Args:
            relationship_type: The type of relationship to create.
            src_entity: Source entity for the relationship.
            dst_entity: Destination entity for the relationship.
            _commit: Optional commit string to tag the relationship.

        Returns:
            List of Records from the transaction.
        """
        self.validate_entities_unique([src_entity, dst_entity])

        rel = R(Type=relationship_type)

        src = N(label=src_entity.get_label(), props=src_entity.get_attr_dict())
        dst = N(label=dst_entity.get_label(), props=dst_entity.get_attr_dict())
        trip = rel.relate(src, dst)
        plain_trip = rel.relate(_plain_var(src), _plain_var(dst))

        try:
            # check for existance shouldn't include _commit?
            self.validate_pattern_unique(trip)
        except self.PatternNotFoundError:
            # if triple doesn't already exist, add _commit? this means that
            # if triple exists w/ different _commit, merge shouldn't add anything?
            rel.props["_commit"] = P(handle="_commit", value=_commit)

        stmt = Statement(Match(src, dst), Merge(plain_trip), use_params=True)

        qry = str(stmt)
        parms = stmt.params

        logger.info(stmt)

        logger.info(
            "Adding %s relationship between src %s with properties: %s to dst %s "
            "with properties: %s",
            relationship_type,
            src.label,
            src_entity.get_attr_dict(),
            dst.label,
            dst_entity.get_attr_dict(),
        )
        return (qry, parms)  # type: ignore[reportReturnType]

    def link_synonyms(
        self,
        entity_1: Entity,
        entity_2: Entity,
        mapping_source: str,
        _commit: str = "",
    ) -> None:
        """
        Link two synonymous entities in the MDB via a Concept node.

        Takes two synonymous Entities (as determined by user/SME) as bento-meta
        objects and connects them to a Concept node via a 'represents' relationship.

        Entities must both exist in the MDB instance and given entity attributes must
        uniquely identify property graph nodes in the MDB.

        Args:
            entity_1: First entity to link.
            entity_2: Second entity to link.
            mapping_source: Source of the mapping relationship.
            _commit: Optional commit string. If set, the _commit property of any node
                created is set to this value.
        """
        self.validate_entities_unique([entity_1, entity_2])

        ent_1_concepts = self.get_concept_nanoids_linked_to_entity(
            entity=entity_1,
            mapping_source=mapping_source,
        )
        ent_2_concepts = self.get_concept_nanoids_linked_to_entity(
            entity=entity_2,
            mapping_source=mapping_source,
        )
        shared_concepts = list(set(ent_1_concepts).intersection(ent_2_concepts))

        # has concept been tagged by this mapping src before
        if shared_concepts:
            logger.warning(
                "This mapping has already been added by this source via "
                "Concept with nanoid: %s",
                shared_concepts[0],
            )
            return

        # one of the entities has a concept created by the given mapping source
        if ent_1_concepts:
            logger.info("Using existing concept with nanoid %s", ent_1_concepts[0])
            concept = Concept({"nanoid": ent_1_concepts[0]})
        elif ent_2_concepts:
            logger.info("Using existing concept with nanoid %s", ent_2_concepts[0])
            concept = Concept({"nanoid": ent_2_concepts[0]})
        else:
            concept = Concept({"nanoid": make_nanoid()})
            self.add_entity_to_mdb(concept, _commit=_commit)
            self.add_tag_to_mdb_entity(
                tag=Tag(
                    {
                        "key": "mapping_source",
                        "value": mapping_source,
                        "nanoid": make_nanoid(),
                    }
                ),
                entity=concept,
            )

        # create specified relationship between each entity and a concept
        rel_type_1 = "represents" if isinstance(entity_1, Term) else "has_concept"
        rel_type_2 = "represents" if isinstance(entity_2, Term) else "has_concept"

        self.add_relationship_to_mdb(
            relationship_type=rel_type_1,
            src_entity=entity_1,
            dst_entity=concept,
            _commit=_commit,
        )
        self.add_relationship_to_mdb(
            relationship_type=rel_type_2,
            src_entity=entity_2,
            dst_entity=concept,
            _commit=_commit,
        )

    @read_txn_value  # type: ignore[reportArgumentType]
    def get_entity_nanoid(self, entity: Entity) -> list[Record]:
        """Take a unique entity in the MDB and return its nanoid."""
        self.validate_entity_unique(entity)

        ent = N(label=entity.get_label(), props=entity.get_attr_dict())

        if isinstance(entity, Edge):
            src = N(label="node", props=entity.src.get_attr_dict())
            dst = N(label="node", props=entity.dst.get_attr_dict())
            src_rel = R(Type="has_src")
            dst_rel = R(Type="has_dst")
            src_trip = src_rel.relate(ent, src)
            dst_trip = dst_rel.relate(ent, dst)
            path = G(src_trip, dst_trip)
            match_clause = Match(path)
        else:
            match_clause = Match(ent)

        stmt = Statement(
            match_clause,
            Return(f"{ent._var}.nanoid"),
            As("entity_nanoid"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "entity_nanoid")  # type: ignore[reportReturnType]

    def get_or_make_entity_nanoid(self, entity: Entity) -> str:
        """Obtain existing entity's nanoid or create one for new entity."""
        try:
            return self.get_entity_nanoid(entity)[0]
        except self.EntityNotFoundError:
            return make_nanoid()

    @read_txn_value  # type: ignore[reportArgumentType]
    def get_term_nanoids(
        self,
        concept: Concept,
        mapping_source: str = "",
    ) -> list[Record]:
        """Return list of term nanoids representing given concept."""
        self.validate_entity_unique(concept)

        ent = N(label="concept", props=concept.get_attr_dict())
        term = N(label="term")
        ent_trip = T(term, R(Type="represents"), ent)

        if mapping_source:
            tag_trip = T(
                concept,
                R(Type="has_tag"),
                N(
                    label="tag",
                    props={"key": "mapping_source", "value": mapping_source},
                ),
            )
            path = G(ent_trip, tag_trip)
        else:
            path = ent_trip

        stmt = Statement(
            Match(path),
            Return(f"{term._var}.nanoid"),
            As("term_nanoids"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "term_nanoids")  # type: ignore[reportReturnType]

    @read_txn_value  # type: ignore[reportArgumentType]
    def get_predicate_nanoids(
        self,
        concept: Concept,
        mapping_source: str = "",
    ) -> list[Record]:
        """Return list of predicate nanoids with relationship to given concept."""
        self.validate_entity_unique(concept)

        ent = N(label="concept", props=concept.get_attr_dict())
        predicate = N(label="predicate")
        ent_trip = T(predicate, R0(), ent)

        if mapping_source:
            tag = N(
                label="tag",
                props={"key": "mapping_source", "value": mapping_source},
            )
            tag_trip = T(ent, R(Type="has_tag"), tag)
            path = G(ent_trip, tag_trip)
        else:
            path = ent_trip

        stmt = Statement(
            Match(path),
            Return(f"{predicate._var}.nanoid"),
            As("predicate_nanoids"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "predicate_nanoids")  # type: ignore[reportReturnType]

    @read_txn_value  # type: ignore[reportArgumentType]
    def get_relationship_between_entities(
        self,
        src_entity: Entity,
        dst_entity: Entity,
    ) -> list[Record]:
        """
        Return relationship type between given entities.

        Relationship pattern: (src)-[:rel_type]->(dst).
        """
        self.validate_entities_unique([src_entity, dst_entity])

        trip = T(
            N(label=src_entity.get_label(), props=src_entity.get_attr_dict()),
            R(),
            N(label=dst_entity.get_label(), props=dst_entity.get_attr_dict()),
        )

        stmt = Statement(
            Match(trip),
            Return(f"TYPE({trip._edge._var})"),
            As("relationship_type"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms, "relationship_type")  # type: ignore[reportReturnType]

    def link_concepts_via_predicate(
        self,
        subject_concept: Concept,
        object_concept: Concept,
        predicate: Predicate,
        _commit: str = "",
    ) -> None:
        """
        Link two synonymous Concepts via a Predicate.

        This function takes two synonymous Concepts as bento-meta objects and links
        them via a Predicate node and has_subject and has_object relationships.

        Args:
            subject_concept: The subject concept.
            object_concept: The object concept.
            predicate: The predicate linking them.
            _commit: Optional commit string to tag created nodes.
        """
        self.validate_entities_unique([subject_concept, object_concept, predicate])

        predicate.subject = subject_concept
        predicate.object = object_concept

        predicate.nanoid = self.get_or_make_entity_nanoid(predicate)
        self.add_entity_to_mdb(predicate, _commit=_commit)

        self.add_relationship_to_mdb(
            relationship_type="has_subject",
            src_entity=predicate,
            dst_entity=predicate.subject,
            _commit=_commit,
        )
        self.add_relationship_to_mdb(
            relationship_type="has_object",
            src_entity=predicate,
            dst_entity=predicate.object,
            _commit=_commit,
        )

    def merge_two_concepts(
        self,
        concept_1: Concept,
        concept_2: Concept,
        mapping_source: str = "",
        _commit: str = "",
    ) -> None:
        """
        Combine two synonymous Concepts into a single Concept.

        This function takes two synonymous Concepts as bento-meta objects and
        merges them into a single Concept along with any connected Terms and Predicates.

        Args:
            concept_1: First concept (will remain after merge).
            concept_2: Second concept (will be deleted after merge).
            mapping_source: Source of the mapping relationship.
            _commit: Optional commit string to tag relationships.
        """
        self.validate_entities_unique([concept_1, concept_2])

        # get list of terms connected to concept 2
        c2_term_nanoids = self.get_term_nanoids(concept_2, mapping_source)
        c2_terms: list[Term] = [Term({"nanoid": nanoid}) for nanoid in c2_term_nanoids]

        # get list of predicates connected to concept 2
        c2_predicate_nanoids = self.get_predicate_nanoids(concept_2, mapping_source)
        c2_predicates_with_rel = []
        for nanoid in c2_predicate_nanoids:
            predicate = Predicate({"nanoid": nanoid})
            predicate_rel = self.get_relationship_between_entities(
                src_entity=predicate,
                dst_entity=concept_2,
            )[0]
            c2_predicates_with_rel.append((predicate, predicate_rel))

        # delete concept 2
        self.remove_entity_from_mdb(concept_2)

        # connect terms from deleted (c2) to remaining concept (c1)
        for term in c2_terms:
            self.add_relationship_to_mdb(
                relationship_type="represents",
                src_entity=term,
                dst_entity=concept_1,
                _commit=_commit,
            )

        # connect predicates from deleted (c2) to remaining concept (c1)
        for predicate, rel in c2_predicates_with_rel:
            self.add_relationship_to_mdb(
                relationship_type=rel,
                src_entity=predicate,
                dst_entity=concept_1,
                _commit=_commit,
            )

    @read_txn_data  # type: ignore[reportArgumentType]
    def _get_all_terms(self) -> list[Record]:
        """Return list of all terms in an MDB."""
        term = N(label="term")

        stmt = Statement(Match(term), Return(term._var))

        qry = str(stmt)
        parms = {}

        logger.debug(qry)

        return (qry, parms)  # type: ignore[reportReturnType]

    def get_potential_term_synonyms(
        self,
        term: Term,
        threshhold: float = 0.8,
    ) -> list[dict]:
        """
        Return list of dicts representing potential synonymous Term nodes.

        Args:
            term: The term to find synonyms for.
            threshhold: Similarity threshold (0-1) for considering terms synonymous.

        Returns:
            List of dicts with value, origin_name, nanoid, similarity, valid_synonym.
        """
        self.validate_entity_unique(term)

        nlp = _get_nlp_model()

        all_terms_result = self._get_all_terms()
        all_terms = [list(item.values())[0] for item in all_terms_result]

        # get likely synonyms
        synonyms = []
        for term_attr_dict in all_terms:
            # calculate similarity between each Term and input Term
            term_1 = nlp(term.value)
            term_2 = nlp(term_attr_dict["value"])
            similarity_score = term_1.similarity(term_2)
            # if similarity threshold met, add to list of potential synonyms
            if similarity_score >= threshhold:
                synonym = {
                    "value": term_attr_dict["value"],
                    "origin_name": term_attr_dict["origin_name"],
                    "nanoid": term_attr_dict["nanoid"],
                    "similarity": similarity_score,
                    "valid_synonym": 0,  # mark 1 if synonym when uploading later
                }
                synonyms.append(synonym)
        return sorted(synonyms, key=lambda d: d["similarity"], reverse=True)

    def potential_synonyms_to_csv(
        self,
        input_data: list[dict],
        output_path: str,
    ) -> None:
        """
        Write CSV with list of synonymous Terms as dicts to given output path.

        Args:
            input_data: List of dicts representing potential synonyms.
            output_path: Path to write CSV file to.
        """
        with Path(output_path).open("w", encoding="utf8", newline="") as output_file:
            dict_writer = csv.DictWriter(output_file, fieldnames=input_data[0].keys())
            dict_writer.writeheader()
            dict_writer.writerows(input_data)

    def link_term_synonyms_csv(
        self,
        term: Term,
        csv_path: str,
        mapping_source: str,
        _commit: str = "",
    ) -> None:
        """
        Link Terms in a CSV of synonymous Terms to given Term via a Concept node.

        Args:
            term: The term to link synonyms to.
            csv_path: Path to CSV file containing potential synonyms.
            mapping_source: Source of the mapping relationship.
            _commit: Optional commit string to tag relationships.
        """
        with Path(csv_path).open(encoding="UTF-8") as csvfile:
            synonym_reader = csv.reader(csvfile)
            for line in synonym_reader:
                if line[3] == "1":  # valid_synonym
                    synonym = Term({"value": line[0], "origin_name": line[1]})
                    self.link_synonyms(
                        entity_1=term,
                        entity_2=synonym,
                        mapping_source=mapping_source,
                        _commit=_commit,
                    )

    @read_txn_data  # type: ignore[reportArgumentType]
    def get_property_synonyms_direct(
        self,
        entity: Property,
        mapping_source: str = "",
    ) -> list[Record]:
        """
        Return list of properties linked by concept to given property.

        Args:
            entity: The property to get synonyms for.
            mapping_source: Optional mapping source to filter by.

        Returns:
            List of Records with property synonyms.
        """
        self.validate_entity_unique(entity)

        ent = N(label="property", props=entity.get_attr_dict())
        prop = N(label="property")
        concept = N(label="concept")
        ent_trip_1 = T(ent, R(Type="has_concept"), concept)
        ent_trip_2 = T(prop, R(Type="has_concept"), concept)

        if mapping_source:
            tag = N(
                label="tag",
                props={"key": "mapping_source", "value": mapping_source},
            )
            tag_trip = T(concept, R(Type="has_tag"), tag)
            path = G(ent_trip_1, ent_trip_2, tag_trip)
        else:
            path = G(ent_trip_1, ent_trip_2)

        stmt = Statement(
            Match(path),
            With(f"{Collect(_plain_var(prop).pattern())}"),
            As("synonyms"),
            Return("synonyms"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms)  # type: ignore[reportReturnType]

    def _get_property_synonyms_direct_as_list(self, entity: Property) -> list[Property]:
        """
        Convert results of read_txn_data-wrapped function with one Property to a list.

        Args:
            entity: The property to get synonyms for.

        Returns:
            List of bento_meta.objects.Property entities.
        """
        data = self.get_property_synonyms_direct(entity)
        return [Property(s) for s in data[0]["synonyms"]]

    def get_property_synonyms_all(self, entity: Property) -> list[Property]:
        """
        Return list of properties linked by concept to given property or its synonyms.

        Chains through synonyms of given property (and so on).

        Args:
            entity: The property to get all synonyms for.

        Returns:
            List of Property entities that are synonyms.
        """
        self.validate_entity_unique(entity)
        all_synonyms = []
        queue = [entity]
        visited = {entity.nanoid}

        while queue:
            current = queue.pop(0)
            direct_synonyms = self._get_property_synonyms_direct_as_list(current)
            for synonym in direct_synonyms:
                if synonym.nanoid not in visited:
                    visited.add(synonym.nanoid)
                    queue.append(synonym)
                    all_synonyms.append(synonym)

        return all_synonyms

    @read_txn_data  # type: ignore[reportArgumentType]
    def _get_property_parents_data(self, entity: Property) -> list[Record]:
        """
        Get list of nodes/edges connected to given property.

        Parents are Nodes or Edges via the "has_property" relationship.

        Args:
            entity: The property to get parents for.

        Returns:
            List of Records with nodes and edges.
        """
        self.validate_entity_unique(entity)
        p_attrs = entity.get_attr_dict()
        child_prop = N(label="property", props=p_attrs)
        parent_node = N(label="node")
        parent_edge = N(label="relationship")
        rel = R(Type="has_property", _dir="_left")
        trip1 = rel.relate(child_prop, N0())
        trip2 = rel.relate(_plain_var(child_prop), parent_node)
        trip3 = rel.relate(_plain_var(child_prop), parent_edge)

        stmt = Statement(
            Match(trip1),
            OptionalMatch(trip2),
            OptionalMatch(trip3),
            With(),
            Collect(_plain_var(parent_node).pattern()),
            As("nodes,"),
            Collect(_plain_var(parent_edge).pattern()),
            As("edges"),
            Return("nodes, edges"),
            use_params=True,
        )

        qry = str(stmt)
        parms = stmt.params

        return (qry, parms)  # type: ignore[reportReturnType]

    def get_property_parents(self, entity: Property) -> list[Node | Edge]:
        """
        Get parents of a property as a list of bento_meta Nodes or Edges.

        Args:
            entity: The property to get parents for.

        Returns:
            List of Node or Edge entities.
        """
        self.validate_entity_unique(entity)

        data = self._get_property_parents_data(entity)
        node_parents = [Node(p) for p in data[0]["nodes"]]
        edge_parents = [Edge(p) for p in data[0]["edges"]]
        return node_parents + edge_parents

    def add_tag_to_mdb_entity(self, tag: Tag, entity: Entity) -> None:
        """
        Add a tag to an existing entity in an MDB.

        Args:
            tag: The tag to add.
            entity: The entity to tag.
        """
        self.validate_entity_unique(entity)
        self.add_entity_to_mdb(tag)
        self.add_relationship_to_mdb(
            relationship_type="has_tag",
            src_entity=entity,
            dst_entity=tag,
        )


class EntityValidator:
    """Class to validate that bento-meta entities have all required attributes."""

    required_attrs_by_entity_type: ClassVar[dict[type[Entity], list[str]]] = {
        Node: ["handle", "model"],
        Edge: ["handle", "model", "src", "dst"],
        Property: ["handle", "model"],
        Term: ["origin_name", "value"],  # add? "origin_id", "origin_version"
        Concept: ["nanoid"],
        Predicate: ["handle", "subject", "object"],
        Tag: ["key", "value"],
        ValueSet: ["handle"],
        Transform: ["handle", "first_step", "last_step", "input_props",
                    "output_props"],
        TfStep: ["package", "entrypoint"],
    }

    valid_attrs: ClassVar[dict[tuple[type[Entity], str], set[str]]] = {
        (Predicate, "handle"): {
            "exactMatch",
            "closeMatch",
            "broader",
            "narrower",
            "related",
        }
    }

    class MissingAttributeError(Exception):
        """An entity doesn't have the attributes required for unique identification."""

    class InvalidAttributeError(Exception):
        """An entity attribute is invalid."""

    @staticmethod
    def validate_entity_has_attribute(entity: Entity, attr_name: str) -> None:
        """Validate the presence of an entity's attribute."""
        if not getattr(entity, attr_name):
            msg = f"{entity.__class__.__name__} needs a {attr_name} attribute"
            raise EntityValidator.MissingAttributeError(msg)

    @staticmethod
    def _validate_entity_attribute(entity_type: type[Entity], attr_name: str) -> None:
        """Check that an entity attribute is in a set of valid attributes."""
        valid_attrs = EntityValidator.valid_attrs.get((entity_type, attr_name))

        if valid_attrs and getattr(entity_type, attr_name) not in valid_attrs:
            msg = (
                f"{entity_type.__name__} {attr_name} must be one of: "
                f"{', '.join(list(valid_attrs))}"
            )
            raise EntityValidator.InvalidAttributeError(msg)

    @staticmethod
    def validate_entity(entity: Entity) -> None:
        """
        Check if entity has all attributes required by MDB for its type.

        Also checks validates all of those attributes are valid themselves if they
        are entities or if they have a fixed set of possible values.

        If looking for a unique identifier for the entity (i.e. nanoid), ensures
        entity has all the required attributes for unique identification.

        Args:
            entity: The entity to validate.

        Raises:
            ValueError: If entity type not supported.
            MissingAttributeError: If required attribute missing.
            InvalidAttributeError: If attribute value invalid.
        """
        entity_type = type(entity)
        required_attrs = EntityValidator.required_attrs_by_entity_type.get(entity_type)

        if not required_attrs:
            msg = f"Entity type {entity_type.__name__} not supported"
            raise ValueError(msg)

        try:
            for attr_name in required_attrs:
                # validate attr that is an entity (e.g. Edge.src)
                if isinstance(getattr(entity, attr_name), Entity):
                    EntityValidator.validate_entity(getattr(entity, attr_name))
                # check entity has attr if required
                EntityValidator.validate_entity_has_attribute(entity, attr_name)
                # check attr with fixed set of possible values is valid
                EntityValidator._validate_entity_attribute(entity_type, attr_name)
        except (
            EntityValidator.MissingAttributeError,
            EntityValidator.InvalidAttributeError,
        ):
            logger.exception("Error validating entity")
            raise



"""
bento_meta.tf_objects

This module contains subclasses of :class:`Entity` that persist and organize
the representation of data transformations between model Properties.
"""
from __future__ import annotations

import json
from typing import TYPE_CHECKING
from .entity import Entity
from .objects import mergespec

if TYPE_CHECKING:
    import neo4j


class Transform(Entity):
    """Subclass that models a data transformation between Properties"""
    attspec_ = {
        "handle": "simple",
        "version": "simple",
        "nanoid": "simple",
        "source": "simple",
        "first_step": "object",
        "last_step": "object",
        "input_props": "collection",
        "output_props": "collection",
        }
    mapspec_ = {
        "label": "transform",
        "key": "handle",
        "property": {
            "handle": "handle",
            "version": "version",
            "nanoid": "nanoid",
            "source": "source",
        },
        "relationship": {
            "input_props": {"rel": "<:value_as_tf_input", "end_cls": "Property"},
            "output_props": {"rel": ":tf_output_as_value", "end_cls": "Property"},
            "first_step": {"rel": "first_tf_step", "end_cls": "TfStep"},
            "last_step": {"rel": "last_tf_step", "end_cls": "TfStep"},
        }
    }
    (attspec, _mapspec) = mergespec("Transform", attspec_, mapspec_)

    def __init__(self, init: dict | neo4j.graph.Node | Transform | None = None) -> None:
        """Initialize a `Transform` instance."""
        super().__init__(init=init)


    @property
    def steps(self):
        ret = []
        s = self.first_step
        while s is not None:
            ret.append(s)
            s = s.next_step
        return ret

class TfStep(Entity):
    """
    Subclass that models a data transformation calculation step in a transform
    workflow
    """
    attspec_ = {
        "package": "simple",
        "version": "simple",
        "entrypoint": "simple",
        "params_json": "simple",
        "input_type": "simple",
        "output_type": "simple",
        "params_type": "simple",
        "next_step": "object",
    }
    mapspec_ = {
        "label": "tf_step",
        "key": "nanoid",
        "property": {
            "package": "package",
            "version": "version",
            "entrypoint": "entrypoint",
            "params_json": "params_json",
            "input_type": "input_type",
            "output_type": "output_type",
            "params_type": "params_type",
            },
        "relationship": {
            "next_step": {"rel": ":next_tf_step>", "end_cls": "TfStep"}
        },
    }
    (attspec, _mapspec) = mergespec("TfSpec", attspec_, mapspec_)
    
    def __init__(self, init: dict | neo4j.graph.Node | TfStep | None = None) -> None:
        """Initialize a `Transform` instance."""
        super().__init__(init=init)

    @property
    def params(self) -> dict | list:
        """
        This property is the python object represented by the JSON string
        contained in 'params_json' (if any)
        """
        if self.params_json is not None:
            return json.loads(self.params_json)
        return



"""
bento_meta.object_map
=====================

This module contains :class:`ObjectMap`, a class which provides the
machinery for mapping bento_meta objects to a Bento Metamodel Database
in Neo4j. Mostly not for human consumption. The ObjectMap:

* interprets the attribute specification (attspec) and map
  specification (mapspec) associated with :class:`Entity` subclasses
* provides the :meth:`get` and :meth:`put` methods to subclasses, that
  enable them to get and put themselves to the database
* generates appropriate
 `Cypher <https://neo4j.com/docs/cypher-manual/current/>`
 queries to do gets and puts
 One ObjectMap instance should be generated for each Entity subclass (see, e.g.,
 :class:`bento_meta.model.Model`)
"""

from __future__ import annotations

import re
import sys

sys.path.append("..")
from typing import Any, ClassVar, cast
from warnings import warn

from neo4j import BoltDriver, Driver, Neo4jDriver, Transaction
from typing_extensions import LiteralString

from bento_meta.entity import ArgError, CollValue, Entity
from bento_meta.objects import (
    Concept,
    Edge,
    Node,
    Origin,
    Predicate,
    Property,
    Tag,
    Term,
    ValueSet,
)
from bento_meta.tf_objects import (
    Transform,
    TfStep,
    )

from pdb import set_trace
class ObjectMap:
    """
    Machinery for mapping bento_meta objects to a Bento Metamodel Database in Neo4j.

    Mostly not for human consumption.
    """

    cache: ClassVar[dict] = {}

    def __init__(
        self,
        *,
        cls: type[Entity] | None = None,
        drv: Driver | None = None,
    ) -> None:
        """
        Initialize the ObjectMap.

        Args:
            cls: The class to map.
            drv: The Neo4j driver.
        """
        if not cls:
            msg = "arg cls= is required"
            raise ArgError(msg)
        self.cls = cls
        if drv:
            if isinstance(drv, (Neo4jDriver, BoltDriver)):
                self.drv = drv
            else:
                msg = (
                    "drv= arg must be Neo4jDriver or BoltDriver "
                    "(returned from GraphDatabase.driver())"
                )
                raise ArgError(msg)
        self.maps = {}

    @classmethod
    def clear_cache(cls) -> None:
        """Clear the cache."""
        cls.cache = {}

    @classmethod
    def cls_by_label(cls, lbl: str) -> type[Entity] | None:
        """Get the class by label."""
        if not hasattr(cls, "_clsxlbl"):
            cls._clsxlbl = {}
            for o in (Node, Edge, Property, ValueSet, Term, Concept, Origin, Tag):
                cls._clsxlbl[o.mapspec()["label"]] = o
        return cls._clsxlbl.get(lbl)

    @classmethod
    def keys_by_cls_and_reln(
        cls,
        qcls: type[Entity],
        reln: str,
    ) -> tuple[str, str | None] | None:
        """Get the keys by class and relationship."""
        if not hasattr(cls, "_keysxcls"):
            cls._keysxcls = {}
            for o in (
                Node,
                Edge,
                Property,
                ValueSet,
                Term,
                Concept,
                Predicate,
                Origin,
                Tag,
                Transform,
                TfStep,
            ):
                for oatt in [x for x in o.attspec if o.attspec[x] == "object"]:
                    r = o.mapspec()["relationship"][oatt]["rel"]
                    r = re.match("[:<>]*([a-zA-Z_]+)[:<>]*", r).group(1)
                    cls._keysxcls[(o.__name__, r)] = (oatt, None)
                for catt in [x for x in o.attspec if o.attspec[x] == "collection"]:
                    r = o.mapspec()["relationship"][catt]["rel"]
                    r = re.match("[:<>]*([a-zA-Z_]+)[:<>]*", r).group(1)
                    cls._keysxcls[(o.__name__, r)] = (catt, o.mapspec()["key"])
        return cls._keysxcls.get((qcls.__name__, reln))

    @classmethod
    def _quote_val(
        cls,
        value: str | float | None,
        *,
        single: bool | None = None,
    ) -> str | float | None:  # double quote unless single is set
        """Quote the value unless single is set."""
        if value is None:
            return None
        if isinstance(value, (int, float)):
            return value  # no quote
        if single:
            return f"'{value}'"  # quote
        return f'"{value}"'  # quote

    def get_by_id(
        self,
        obj: Entity,
        id: str,
        *,
        refresh: bool = False,
    ) -> Entity | None:
        """Get an entity given an id attribute value (not the Neo4j id)."""
        neoid = None

        if self.drv is None:
            msg = "get_by_id() requires Neo4j driver instance"
            raise ArgError(msg)

        with self.drv.session() as session:
            result = session.run(cast("LiteralString", self.get_by_id_q()), {"id": id})
            rec = (
                result.single()
            )  # should be unique - this call will warn if there are more than one

            if rec is not None:
                neoid = rec["id(n)"]

        if neoid is not None:
            obj.neoid = neoid
            return self.get(obj, refresh=True)
        return None

    def get_by_node_nanoid(
        self,
        obj: Entity,
        nanoid: str,
        *,
        refresh: bool = False,
    ) -> Entity | None:
        """PROTOTYPE: Get an entity given an id attribute value (not the Neo4j id)."""
        neo4jid = None

        if not self.drv:
            msg = "get_by_id() requires Neo4j driver instance"
            raise ArgError(msg)

        with self.drv.session() as session:
            result = session.run(
                cast("LiteralString", self.get_by_node_nanoid_q()),
                {"nanoid": nanoid},
            )
            rec = (
                result.single()
            )  # should be unique - this call will warn if there are more than one
            if rec is not None:
                neo4jid = rec["id(n)"]

        if neo4jid is None:
            obj.neoid = neo4jid
            return self.get(obj, refresh=True)
        return None

    def get(self, obj: Entity, *, refresh: bool = False) -> Entity:
        """Get the data for an object instance from the db and load the instance with it."""
        if not self.drv:
            msg = "get() requires Neo4j driver instance"
            raise ArgError(msg)

        if refresh:
            pass
        elif (obj.neoid in ObjectMap.cache) and (ObjectMap.cache[obj.neoid].dirty >= 0):
            return obj

        with self.drv.session() as session:
            result = session.run(cast("LiteralString", self.get_q(obj)))
            rec = result.single()
            if not rec:
                msg = f"object with id {obj.neoid} not found in db"
                raise RuntimeError(msg)

        if obj.neoid not in ObjectMap.cache:
            ObjectMap.cache[obj.neoid] = obj

        with self.drv.session() as session:
            for att in self.cls.mapspec()["relationship"]:
                result = session.run(cast("LiteralString", self.get_attr_q(obj, att)))
                values = {}
                first_val = None
                for rec in result:
                    o = ObjectMap.cache.get(rec["a"].id)
                    if o:
                        if not first_val:
                            first_val = o
                        values[getattr(o, type(o).mapspec()["key"])] = o
                    else:
                        c = None
                        for lbl in rec["a"].labels:
                            c = ObjectMap.cls_by_label(lbl)
                            if c:
                                break
                        if not c:
                            msg = (
                                f"node labels {rec['a'].labels} "
                                "have no associated class in the object model"
                            )
                            raise RuntimeError(msg)
                        o = c(rec["a"])
                        o.dirty = -1
                        ObjectMap.cache[o.neoid] = o
                        if not first_val:
                            first_val = o
                        values[getattr(o, type(o).mapspec()["key"])] = o
                if self.cls.attspec[att] == "object" and len(values) > 1:
                    warn(
                        (
                            f"expected one node for attribute {att} on class "
                            f"{self.cls.__name__}, but got {len(values)}; using first one"
                        ),
                        stacklevel=2,
                    )
                if self.cls.attspec[att] == "object":
                    setattr(obj, att, first_val)
                elif self.cls.attspec[att] == "collection":
                    setattr(obj, att, values)
                else:
                    msg = (
                        f"attribute '{att}' has unknown attribute type "
                        f"'{self.cls.attspec[att]}'"
                    )
                    raise RuntimeError(msg)

        obj.clear_removed_entities()
        obj.dirty = 0
        return obj

    def put(self, obj: Entity) -> Entity:
        """Put the object instance's attributes to the mapped data node in the database."""
        if not self.drv:
            msg = "put() requires Neo4j driver instance"
            raise ArgError(msg)
        with self.drv.session() as session:
            result = None
            with session.begin_transaction() as tx:
                for qry in self.put_q(obj):
                    result = tx.run(cast("LiteralString", qry[0]), qry[1])
                if result is None:
                    msg = "no result from put_q"
                    raise RuntimeError(msg)
                obj.neoid = result.single().value("id(n)")
                if obj.neoid is None:
                    msg = (
                        "no neo4j id retrived on put for obj "
                        f"'{getattr(obj, self.cls.mapspec()['key'])}'"
                    )
                    raise RuntimeError(msg)
                for att in self.cls.mapspec()["relationship"]:
                    values = getattr(obj, att)
                    if not values:
                        continue
                    if isinstance(values, CollValue):
                        items = values.values()
                    else:
                        items = [values]
                    for val in items:
                        if val.neoid is not None:
                            continue
                        # put val as a node
                        for qry in ObjectMap(cls=type(val), drv=self.drv).put_q(val):
                            result = tx.run(cast("LiteralString", qry[0]),qry[1])
                        val.neoid = result.single().value("id(n)")
                        if val.neoid is None:
                            msg = (
                                "no neo4j id retrived on put for obj "
                                f"'{val[type(val).mapspec()['key']]}'"
                            )
                            raise RuntimeError(msg)
                        val.dirty = 1
                        ObjectMap.cache[val.neoid] = val
                    for qry in self.put_attr_q(obj, att, values):
                        tx.run(cast("LiteralString", qry))
                    # drop removed entities here
                    while obj.removed_entities:
                        ent = obj.removed_entities.pop()
                        self.drop(obj, *ent, tx)
        ObjectMap.cache[obj.neoid] = obj
        obj.dirty = 0
        return obj

    def rm(self, obj: Entity, *, force: bool | int = False) -> Any | None:
        """'Delete' the object's mapped node from the database."""
        if not self.drv:
            msg = "rm() requires Neo4j driver instance"
            raise ArgError(msg)
        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        with self.drv.session() as session:
            result = session.run(cast("LiteralString", self.rm_q(obj, detach=force)))
            s = result.single()
            if s is None:
                warn("rm() - corresponding db node not found", stacklevel=2)
            else:
                return s.value()
        return None

    def add(self, obj: Entity, att: str, tgt: Entity) -> Any:
        """
        Create a link between an object instance and a target object in the database.

        This represents adding an object-valued attribute to the object.

        Args:
            obj: The object instance to add attribute to.
            att: The attribute name.
            tgt: The target entity to link.

        Returns:
            The Neo4j ID of the target, or None if not found.
        """
        if not self.drv:
            msg = "add() requires Neo4j driver instance"
            raise ArgError(msg)
        with self.drv.session() as session:
            for qry in self.put_attr_q(obj, att, tgt):
                result = session.run(cast("LiteralString", qry))
            tgt_id = result.single().value()
            if tgt_id is None:
                warn("add() - corresponding db node not found", stacklevel=2)
            return tgt_id

    def drop(
        self,
        obj: Entity,
        att: str,
        tgt: Entity,
        tx: Transaction | None = None,
    ) -> Any:
        """
        Remove an existing link between an object instance and a target object in the database.

        This represents dropping an object-valued attribute from the object.

        Args:
            obj: The object instance to remove attribute from.
            att: The attribute name.
            tgt: The target entity to unlink.
            tx: Optional transaction to use for the operation.

        Returns:
            The result value, or None if not found.
        """
        if not self.drv:
            msg = "rm() requires Neo4j driver instance"
            raise ArgError(msg)
        # if the tgt is not in the database, then dropping it is a no-op:
        if not tgt.neoid:
            return None

        if tx:
            result = None
            for qry in self.rm_attr_q(obj, att, tgt):
                result = tx.run(cast("LiteralString", qry))
            s = result.single()
            if s is None:
                warn("drop() - corresponding target db node not found", stacklevel=2)
            else:
                return s.value()
        else:
            with self.drv.session() as session:
                result = None
                for qry in self.rm_attr_q(obj, att, tgt):
                    result = session.run(cast("LiteralString", qry))
                s = result.single()
                if s is None:
                    warn(
                        "drop() - corresponding target db node not found",
                        stacklevel=2,
                    )
                else:
                    return s.value()
        return None

    def get_owners(
        self,
        obj: Entity,
    ) -> list[tuple[Entity, tuple[str, str | None] | None]]:
        """Get the nodes which are linked to the object instance (the owners of the object)."""
        if not self.drv:
            msg = "get_owners() requires Neo4j driver instance"
            raise ArgError(msg)
        ret = []
        with self.drv.session() as session:
            result = session.run(cast("LiteralString", self.get_owners_q(obj)))
            for rec in result:
                if rec["reln"][0] == "_":  # skip _prev, _next, and convenience links
                    break
                ocls = None
                for lbl in rec["a"].labels:
                    ocls = self.cls_by_label(lbl)
                    if ocls:
                        break
                assert ocls
                o = ocls(rec["a"])
                # creating object but not putting in the cache - why?
                keys = self.keys_by_cls_and_reln(type(o), rec["reln"])
                # obj.belongs[(id(o),*keys)] = o
                # not setting the belongs on the obj - why?
                ret.append((o, keys))
        return ret

    def get_q(self, obj: Entity) -> str:
        """Get the query for an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)
        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        return (
            f"MATCH (n:{self.cls.mapspec()['label']}) "
            f"WHERE id(n)={obj.neoid} RETURN n,id(n)"
        )

    def get_by_id_q(self) -> str:
        """Get the query for an entity by its Neo4j id."""
        return (
            f"MATCH (n:{self.cls.mapspec()['label']}) "
            "WHERE id(n)=$id and n._to IS NULL RETURN id(n)"
        )

    def get_by_node_nanoid_q(self) -> str:
        """PROTOTYPE: Get the query for an entity given its nanoid."""
        return "MATCH (n:node) WHERE n.nanoid=$nanoid and n._to is NULL RETURN id(n)"

    def get_attr_q(self, obj: Entity, att: str) -> str:
        """Get the query for an attribute of an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)
        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        label = self.cls.mapspec()["label"]
        if att in self.cls.mapspec()["property"]:
            pr = self.cls.mapspec()["property"][att]
            return f"MATCH (n:{label}) WHERE id(n)={obj.neoid} RETURN n.{pr}"
        if att in self.cls.mapspec()["relationship"]:
            spec = self.cls.mapspec()["relationship"][att]
            end_cls = spec["end_cls"]
            if isinstance(end_cls, str):
                end_cls = {end_cls}
            end_lbls = [eval(x).mapspec()["label"] for x in end_cls]
            rel = re.sub("^([^:]?)(:[a-zA-Z0-9_]+)(.*)$", r"\1-[\2]-\3", spec["rel"])
            if len(end_lbls) == 1:
                qry = (
                    f"MATCH (n:{label}){rel}(a:{end_lbls[0]}) "
                    f"WHERE id(n)={obj.neoid} RETURN a"
                )
                if self.cls.attspec[att] == "object":
                    qry += " LIMIT 1"
                return qry
            # multiple end classes possible
            cond = " OR ".join([f"'{lbl}' IN labels(a)" for lbl in end_lbls])
            return (
                f"MATCH (n:{label}){rel}(a) WHERE id(n)={obj.neoid} AND ({cond}) "
                "RETURN a"
            )
        msg = f"'{att}' is not a registered attribute for class '{self.cls.__name__}'"
        raise ArgError(msg)

    def get_owners_q(self, obj: Entity) -> str:
        """Get the query for the owners of an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)

        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        label = self.cls.mapspec()["label"]
        return (
            f"MATCH (n:{label})<-[r]-(a) WHERE id(n)={obj.neoid} "
            "RETURN TYPE(r) as reln, a"
        )

    def put_q(self, obj: Entity) -> list[str]:
        """Get the query for putting an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)
        props = {}
        null_props = []
        for pr in self.cls.mapspec()["property"]:
            if getattr(obj, pr) is None:
                null_props.append(self.cls.mapspec()["property"][pr])
            else:
                props[self.cls.mapspec()["property"][pr]] = getattr(obj, pr)
        stmts = []
        if obj.neoid is not None:
            i = 0
            prms = {}
            assigns = []
            for pr in props:
                assigns.append(f"n.{pr}=$p{i}")
                prms[f"p{i}"] = props[pr]
                i = i + 1
            set_clause = "SET " + ",".join(assigns)
            stmts.append(
                (f"MATCH (n:{self.cls.mapspec()['label']}) WHERE id(n)={obj.neoid} "
                 f"{set_clause} RETURN n,id(n)", prms)
            )
            rmstmts = []
            for pr in null_props:
                rmstmts.append(
                    ((f"MATCH (n:{self.cls.mapspec()['label']}) WHERE id(n)="
                      f"{obj.neoid} REMOVE n.{pr} RETURN n,id(n)"), {})
                )
            stmts.extend(rmstmts)
            return stmts
        else:
            i = 0
            prms = {}
            maps = []
            for pr in props:
                maps.append(f"{pr}:$p{i}")
                prms[f"p{i}"] = props[pr]
                i = i + 1
            spec = ",".join(maps)
            return [
                (f"CREATE (n:{self.cls.mapspec()['label']} {{{spec}}}) RETURN n,id(n)", prms),
            ]

    def put_attr_q(
        self,
        obj: Entity,
        att: str,
        values: Entity | list[Entity] | CollValue,
    ) -> str | list[str]:
        """Get the query for putting an attribute of an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)
        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        if not isinstance(values, (Entity, list, CollValue)):
            msg = "'values' must be a list of values suitable for the attribute"
            raise ArgError(msg)
        if isinstance(values, CollValue):
            values = values.values()
        elif isinstance(values, Entity):
            values = [values]
        if att in self.cls.mapspec()["property"]:
            return (
                f"MATCH (n:{self.cls.mapspec()['label']}) WHERE id(n)={obj.neoid} "
                f"SET {self.cls.mapspec()['property'][att]}="
                f"{ObjectMap._quote_val(values[0])} RETURN id(n)"
            )
        if att in self.cls.mapspec()["relationship"]:
            if not self._check_values_list(att, values):
                msg = (
                    "'values' must be a list of mapped Entity objects of "
                    f"the appropriate subclass for attribute '{att}'",
                )
                raise ArgError(msg)
            stmts = []
            spec = self.cls.mapspec()["relationship"][att]
            end_cls = spec["end_cls"]
            if isinstance(end_cls, str):
                end_cls = {end_cls}
            end_lbls = [eval(x).mapspec()["label"] for x in end_cls]
            rel = re.sub("^([^:]?)(:[a-zA-Z0-9_]+)(.*)$", r"\1-[\2]-\3", spec["rel"])
            cond = " OR ".join([f"'{lbl}' IN labels(a)" for lbl in end_lbls])
            for avalue in values:
                if len(end_lbls) == 1:
                    stmts.append(
                        f"MATCH (n:{self.cls.mapspec()['label']}),(a:{end_lbls[0]}) "
                        f"WHERE id(n)={obj.neoid} AND id(a)={avalue.neoid} "
                        f"MERGE (n){rel}(a) RETURN id(a)",
                    )
                else:
                    stmts.append(
                        f"MATCH (n:{self.cls.mapspec()['label']}),(a) "
                        f"WHERE id(n)={obj.neoid} AND id(a)={avalue.neoid} AND "
                        f"({cond}) MERGE (n){rel}(a) RETURN id(a)",
                    )
            return stmts
        msg = f"'{att}' is not a registered attribute for class '{self.cls.__name__}'"
        raise ArgError(msg)

    def rm_q(self, obj: Entity, *, detach: bool = False) -> str:
        """Get the query for removing an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)
        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        dlt = "DETACH DELETE n" if detach else "DELETE n"
        qry = f"MATCH (n:{self.cls.mapspec()['label']}) WHERE id(n)={obj.neoid} "
        return qry + dlt

    def rm_attr_q(
        self,
        obj: Entity,
        att: str,
        values: list[Entity] | None = None,
    ) -> str | list[str]:
        """Get the query for removing an attribute of an object."""
        if not isinstance(obj, self.cls):
            msg = f"arg1 must be object of class {self.cls.__name__}"
            raise ArgError(msg)
        if obj.neoid is None:
            msg = "object must be mapped (i.e., obj.neoid must be set)"
            raise ArgError(msg)
        if values and not isinstance(values, list):
            values = [values]
        if att in self.cls.mapspec()["property"]:
            return (
                f"MATCH (n:{self.cls.mapspec()['label']}) "
                f"WHERE id(n)={obj.neoid} REMOVE n.{att} RETURN id(n)"
            )
        if att in self.cls.mapspec()["relationship"]:
            many = self.cls.attspec[att] == "collection"
            spec = self.cls.mapspec()["relationship"][att]
            end_cls = spec["end_cls"]
            if isinstance(end_cls, str):
                end_cls = {end_cls}
            end_lbls = [eval(x).mapspec()["label"] for x in end_cls]
            cond = " OR ".join([f"'{lbl}' IN labels(a)" for lbl in end_lbls])
            rel = re.sub("^([^:]?)(:[a-zA-Z0-9_]+)(.*)$", r"\1-[r\2]-\3", spec["rel"])
            if values and values[0] == ":all":
                if len(end_lbls) == 1:
                    return (
                        f"MATCH (n:{self.cls.mapspec()['label']}){rel}(a:{end_lbls[0]})"
                        f" WHERE id(n)={obj.neoid} DELETE r RETURN id(n),id(a)"
                    )
                return (
                    f"MATCH (n:{self.cls.mapspec()['label']}){rel}(a) "
                    f"WHERE id(n)={obj.neoid} AND ({cond}) DELETE r RETURN id(n)"
                )
            stmts = []

            if not self._check_values_list(att, values):
                msg = (
                    "'values' must be a list of mapped Entity objects of the "
                    f"appropriate subclass for attribute '{att}'",
                )
                raise ArgError(msg)
            for val in values:
                qry = ""
                if len(end_lbls) == 1:
                    qry = (
                        f"MATCH (n:{self.cls.mapspec()['label']}){rel}(a:{end_lbls[0]})"
                        f" WHERE id(n)={obj.neoid} AND id(a)={val.neoid} "
                        f"DELETE r RETURN id(n),id(a)"
                    )
                else:
                    qry = (
                        f"MATCH (n:{self.cls.mapspec()['label']}){rel}(a) "
                        f"WHERE id(n)={obj.neoid} AND id(a)={val.neoid} AND ({cond}) "
                        f"DELETE r RETURN id(n),id(a)"
                    )
                stmts.append(qry)
            return stmts
        msg = f"'{att}' is not a registered attribute for class '{self.cls.__name__}'"
        raise ArgError(msg)

    def _check_values_list(self, att: str, values: list[Entity] | CollValue) -> bool:
        """Check if the values are a list of mapped Entity objects of the appropriate subclass for an attribute."""
        v = values
        if isinstance(values, CollValue):
            v = values.values()
        chk = [x.neoid is None for x in v]
        if True in chk:
            return False
        end_cls = self.cls.mapspec()["relationship"][att]["end_cls"]
        if isinstance(end_cls, str):
            end_cls = {end_cls}
        cls_set = tuple([eval(x) for x in end_cls])
        # print(f"{cls_set=}")
        # print(f"{v=}")
        chk = [isinstance(x, cls_set) for x in v]
        return True in chk



"""
mdb.searchable
Subclass of `class:bento_meta.mdb.MDB` to support searching fulltext indices on an MDB
Note: certain fulltext indexes on certain MDB nodes and properties must be present in
the Neo4j instance:
- nodeHandle
- propHandle
- termValue
- termDefn
- termValueDefn

"""

from typing import Any

from bento_meta.mdb import MDB, read_txn_data


class SearchableMDB(MDB):
    """:class:`bento_meta.mdb.MDB` subclass for searching fulltext indices on an MDB."""

    def __init__(
        self,
        uri: str | None = None,
        user: str | None = None,
        password: str | None = None,
    ) -> None:
        """Initialize a :class:`SearchableMDB` object."""
        super().__init__(uri=uri, user=user, password=password)
        self.ftindexes = {}
        with self.driver.session() as s:
            result = s.run("call db.indexes")
            for rec in result:
                if rec["type"] == "FULLTEXT":
                    self.ftindexes[rec["name"]] = {
                        "entity_type": rec["entityType"],
                        "entities": rec["labelsOrTypes"],
                        "properties": rec["properties"],
                    }

    def available_indexes(self) -> dict[str, dict[str, list[str]]]:
        """
        Fulltext indexes present in database.

        Returns:
            Dict mapping index_name to dict with entity_type (NODE|RELATIONSHIP),
            entities ([labels]), properties ([[props]]).
        """
        return self.ftindexes

    @read_txn_data  # type: ignore[reportArgumentType]
    def query_index(
        self,
        index: str,
        qstring: str,
        skip: str | None = None,
        limit: str | None = None,
    ) -> list[dict[str, Any]] | None:
        """
        Query a named fulltext index of nodes or relationships.

        Args:
            index: Name of the fulltext index to query.
            qstring: Lucene query string.
            skip: Number of results to skip.
            limit: Maximum number of results to return.

        Returns:
            List of dicts with ent (entity dict), label, score (lucene score).
        """
        if index not in self.ftindexes:
            msg = f"Index with name '{index}' not found"
            raise RuntimeError(msg)
        tipe = ""
        if self.ftindexes[index]["entity_type"] == "NODE":
            tipe = "queryNodes"
        elif self.ftindexes[index]["entity_type"] == "RELATIONSHIP":
            tipe = "queryRelationships"
        else:
            msg = "Wha?"
            raise RuntimeError(msg)
        thing = self.ftindexes[index]["entity_type"].lower()
        qry = (
            "call db.index.fulltext.{tipe}($name, $query) "
            "yield {thing}, score "
            "return {thing} as ent, head(labels({thing})) as label, score "
            "{skip} {limit} "
        ).format(
            tipe=tipe,
            thing=thing,
            skip="SKIP $skip" if skip else "",
            limit="LIMIT $limit" if limit else "",
        )
        parms = {"name": index, "query": qstring}
        if skip:
            parms["skip"] = skip
        if limit:
            parms["limit"] = limit
        return (qry, parms)  # type: ignore[reportReturnType]

    def search_entity_handles(
        self,
        qstring: str,
    ) -> dict[str, list[dict[str, Any]]] | None:
        """
        Fulltext search of qstring over node, relationship, and property handles.

        Args:
            qstring: Lucene query string.

        Returns:
            Dict with nodes, relationships, properties, each containing list of dicts
            with ent (entity dict) and score (lucene score).
        """
        result = self.query_index("entityHandle", qstring)
        if not result:
            return None
        plural = {
            "node": "nodes",
            "relationship": "relationships",
            "property": "properties",
        }
        ret = {"nodes": [], "relationships": [], "properties": []}
        for item in result:
            ret[plural[item["label"]]].append(
                {"ent": item["ent"], "score": item["score"]}
            )
        return ret

    def search_terms(
        self,
        qstring: str,
        *,
        search_values: bool = True,
        search_definitions: bool = True,
    ) -> list[dict[str, Any]] | None:
        """
        Fulltext search for qstring over terms, by value, definition, or both (default).

        Args:
            qstring: Lucene query string.
            search_values: If True, search term values.
            search_definitions: If True, search term definitions.

        Returns:
            List of dicts with ent (term dict) and score (lucene score).
        """
        index = {
            True: {True: "termValueDefn", False: "termDefn"},
            False: {True: "termValue", False: None},
        }
        return self.query_index(index[search_definitions][search_values], qstring)



"""
makeq - make a Neo4j query from an endpoint path.
"""

import re

import yaml
from minicypher.functions import (
    And,
    Not,
    Or,
    count,
    exists,
    group,
    labels,
)

from bento_meta.util._engine import _engine

avail_funcs = {x.__name__: x for x in (count, exists, labels, group, And, Or, Not)}


def f(pfx, pth):
    tok = [x for x in pth if x.startswith("$")]
    if not tok:
        tok = [x for x in pth if not x.startswith("_")]
    if not tok:
        print(pfx)
        return
    if pth.get("_return"):
        print(pfx)
    for t in tok:
        f("/".join([pfx, t]), pth[t])
    return


class Query:
    paths = {}
    cache = {}

    def __init__(self, path, use_cache=True):
        path = path.removeprefix("/")
        self.toks = path.split("/")
        self._engine = None
        if use_cache:
            # interpret the cache key as a regexp matching the input path
            hit = [x for x in self.cache if re.match("^" + x + "$", path)]
            if hit:
                Q = self.cache[hit[0]]
                # pull the new parameter values from the path
                vals = re.match(hit[0], path).groups()
                keys = sorted(Q._engine.params.keys())
                self._engine = Q._engine
                for pr in zip(keys, vals):
                    self._engine.params[pr[0]] = pr[1]
        if not self._engine:
            self._engine = _engine()
            if not self._engine.parse(self.toks):
                raise RuntimeError(self._engine.error)
            if use_cache:
                self.cache[self._engine.key] = self

    @classmethod
    def set_paths(cls, paths):
        if paths.get("paths"):
            cls.paths = paths["paths"]
        else:
            cls.paths = paths
        _engine.set_paths(cls.paths)
        return True

    @classmethod
    def load_paths(cls, flo):
        p = yaml.load(flo, Loader=yaml.CLoader)
        return cls.set_paths(p)

    @property
    def statement(self):
        return self._engine.statement

    @property
    def params(self):
        return self._engine.params

    @property
    def path_id(self):
        return self._engine.path_id

    def __str__(self):
        return str(self.statement)
